r"""Neural networks, layers and modules."""

__all__ = ["MLP", "Linear", "MaskedMLP", "MonotonicMLP"]

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import BoolTensor, Tensor
from typing import Callable, Iterable, Sequence, Union


def linear(x: Tensor, W: Tensor, b: Tensor = None) -> Tensor:
    if W.dim() == 2:
        return F.linear(x, W, b)
    else:
        x = torch.einsum("...ij,...j->...i", W, x)

    if b is None:
        return x
    else:
        return x + b


class LayerNorm(nn.Module):
    r"""Creates a normalization layer that standardizes features along a dimension.

    .. math:: y = \frac{x - \mathrm{E}[x]}{\sqrt{\mathrm{Var}[x] + \epsilon}}

    References:
       Layer Normalization (Lei Ba et al., 2016)
       https://arxiv.org/abs/1607.06450

    Arguments:
        dim: The dimension(s) to standardize.
        eps: A numerical stability term.
    """

    def __init__(self, dim: Union[int, Iterable[int]] = -1, eps: float = 1e-5):
        super().__init__()

        self.dim = dim if isinstance(dim, int) else tuple(dim)
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        variance, mean = torch.var_mean(x, unbiased=True, dim=self.dim, keepdim=True)

        return (x - mean) / (variance + self.eps).sqrt()


class Linear(nn.Module):
    r"""Creates a linear layer.

    .. math:: y = x W^T + b

    If the :py:`stack` argument is provided, the module creates a stack of
    independent linear operators that are applied to a stack of input vectors.

    Arguments:
        in_features: The number of input features :math:`C`.
        out_features: The number of output features :math:`C'`.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        stack: The number of stacked operators :math:`S`.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        stack: int = None,
    ):
        super().__init__()

        shape = () if stack is None else (stack,)

        self.weight = nn.Parameter(torch.empty(*shape, out_features, in_features))

        if bias:
            self.bias = nn.Parameter(torch.empty(*shape, out_features))
        else:
            self.bias = None

        self.reset_parameters()

        self.in_features = in_features
        self.out_features = out_features

    def reset_parameters(self):
        bound = 1 / self.weight.shape[-1] ** 0.5

        nn.init.uniform_(self.weight, -bound, bound)

        if self.bias is not None:
            nn.init.uniform_(self.bias, -bound, bound)

    def extra_repr(self) -> str:
        if self.weight.dim() == 2:
            stack, fout, fin = (None, *self.weight.shape)
        else:
            stack, fout, fin = self.weight.shape

        bias = self.bias is not None

        if stack is None:
            return f"in_features={fin}, out_features={fout}, bias={bias}"
        else:
            return f"in_features={fin}, out_features={fout}, bias={bias}, stack={stack}"

    def forward(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, C)` or :math:`(*, S, C)`.

        Returns:
            The output tensor :math:`y`, with shape or :math:`(*, C')` or :math:`(*, S, C')`.
        """

        return linear(x, self.weight, self.bias)


class MLP(nn.Sequential):
    r"""Creates a multi-layer perceptron (MLP).

    Also known as fully connected feedforward network, an MLP is a sequence of
    non-linear parametric functions

    .. math:: h_{i + 1} = a_{i + 1}(h_i W_{i + 1}^T + b_{i + 1}),

    over feature vectors :math:`h_i`, with the input and output feature vectors
    :math:`x = h_0` and :math:`y = h_L`, respectively. The non-linear functions
    :math:`a_i` are called activation functions. The trainable parameters of an MLP
    are its weights and biases :math:`\phi = \{W_i, b_i | i = 1, \dots, L\}`.

    Wikipedia:
        https://wikipedia.org/wiki/Feedforward_neural_network

    Arguments:
        in_features: The number of input features.
        out_features: The number of output features.
        hidden_features: The numbers of hidden features.
        activation: The activation function constructor. If :py:`None`, use
            :class:`torch.nn.ReLU` instead.
        normalize: Whether features are normalized between layers or not.
        kwargs: Keyword arguments passed to :class:`Linear`.

    Example:
        >>> net = MLP(64, 1, [32, 16], activation=nn.ELU)
        >>> net
        MLP(
          (0): Linear(in_features=64, out_features=32, bias=True)
          (1): ELU(alpha=1.0)
          (2): Linear(in_features=32, out_features=16, bias=True)
          (3): ELU(alpha=1.0)
          (4): Linear(in_features=16, out_features=1, bias=True)
        )
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_features: Sequence[int] = (64, 64),
        activation: Callable[[], nn.Module] = None,
        normalize: bool = False,
        **kwargs,
    ):
        if activation is None:
            activation = nn.ReLU

        normalization = LayerNorm if normalize else lambda: None

        layers = []

        for before, after in zip(
            (in_features, *hidden_features),
            (*hidden_features, out_features),
        ):
            layers.extend([
                Linear(before, after, **kwargs),
                activation(),
                normalization(),
            ])

        layers = layers[:-2]
        layers = filter(lambda layer: layer is not None, layers)

        super().__init__(*layers)

        self.in_features = in_features
        self.out_features = out_features


class Residual(nn.Sequential):
    r"""Creates a residual block."""

    def forward(self, x: Tensor) -> Tensor:
        return x + super().forward(x)


class MaskedLinear(nn.Linear):
    r"""Creates a masked linear layer.

    .. math:: y = x (W \odot A)^T + b

    Arguments:
        adjacency: The adjacency matrix :math:`A \in \{0, 1\}^{M \times N}`.
        kwargs: Keyword arguments passed to :class:`torch.nn.Linear`.
    """

    def __init__(self, adjacency: BoolTensor, **kwargs):
        super().__init__(*reversed(adjacency.shape), **kwargs)

        self.register_buffer("mask", adjacency)

    def forward(self, x: Tensor) -> Tensor:
        return F.linear(x, self.mask * self.weight, self.bias)


class MaskedMLP(nn.Sequential):
    r"""Creates a masked multi-layer perceptron (MaskedMLP).

    The resulting MLP is a transformation :math:`y = f(x)` whose Jacobian entries
    :math:`\frac{\partial y_i}{\partial x_j}` are null if :math:`A_{ij} = 0`.

    Arguments:
        adjacency: The adjacency matrix :math:`A \in \{0, 1\}^{M \times N}`.
        hidden_features: The numbers of hidden features.
        activation: The activation function constructor. If :py:`None`, use
            :class:`torch.nn.ReLU` instead.
        residual: Whether to use residual blocks or not.

    Example:
        >>> adjacency = torch.randn(4, 3) < 0
        >>> adjacency
        tensor([[False,  True,  True],
                [False,  True,  True],
                [False, False,  True],
                [ True,  True, False]])
        >>> net = MaskedMLP(adjacency, [16, 32], activation=nn.ELU)
        >>> net
        MaskedMLP(
          (0): MaskedLinear(in_features=3, out_features=16, bias=True)
          (1): ELU(alpha=1.0)
          (2): MaskedLinear(in_features=16, out_features=32, bias=True)
          (3): ELU(alpha=1.0)
          (4): MaskedLinear(in_features=32, out_features=4, bias=True)
        )
        >>> x = torch.randn(3)
        >>> torch.autograd.functional.jacobian(net, x)
        tensor([[ 0.0000, -0.0065,  0.1158],
                [ 0.0000, -0.0089,  0.0072],
                [ 0.0000,  0.0000,  0.0089],
                [-0.0146, -0.0128,  0.0000]])
    """

    def __init__(
        self,
        adjacency: BoolTensor,
        hidden_features: Sequence[int] = (64, 64),
        activation: Callable[[], nn.Module] = None,
        residual: bool = False,
    ):
        out_features, in_features = adjacency.shape

        if activation is None:
            activation = nn.ReLU

        # Merge outputs with the same dependencies
        adjacency, inverse = torch.unique(adjacency, dim=0, return_inverse=True)

        # P_ij = 1 if A_ik = 1 for all k such that A_jk = 1
        precedence = adjacency.double() @ adjacency.double().t() == adjacency.sum(dim=-1)

        # Layers
        layers = []

        for i, features in enumerate((*hidden_features, out_features)):
            if i > 0:
                mask = precedence[:, indices]  # noqa: F821
            else:
                mask = adjacency

            if (~mask).all():
                raise ValueError("The adjacency matrix leads to a null Jacobian.")

            if i < len(hidden_features):
                reachable = mask.sum(dim=-1).nonzero().squeeze(dim=-1)
                indices = reachable[torch.arange(features) % len(reachable)]
                mask = mask[indices]
            else:
                mask = mask[inverse]

            layers.append(MaskedLinear(adjacency=mask))

            if residual:
                if 0 < i < len(hidden_features) and mask.shape[0] == mask.shape[1]:
                    layers.pop()

                mask = precedence[indices, :][:, indices]

                layers.append(
                    Residual(
                        MaskedLinear(adjacency=mask),
                        activation(),
                        MaskedLinear(adjacency=mask),
                    )
                )
            else:
                layers.append(activation())

        layers.pop()

        super().__init__(*layers)

        self.in_features = in_features
        self.out_features = out_features


class MonotonicLinear(Linear):
    r"""Creates a monotonic linear layer.

    .. math:: y = x |W|^T + b

    Arguments:
        args: Positional arguments passed to :class:`Linear`.
        kwargs: Keyword arguments passed to :class:`Linear`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return linear(x, self.weight.abs(), self.bias)


class TwoWayELU(nn.ELU):
    r"""Creates a layer that splits the input into two groups and applies
    :math:`\text{ELU}(x)` to the first and :math:`-\text{ELU}(-x)` to the second.

    Arguments:
        args: Positional arguments passed to :class:`torch.nn.ELU`.
        kwargs: Keyword arguments passed to :class:`torch.nn.ELU`.
    """

    def forward(self, x: Tensor) -> Tensor:
        x0, x1 = torch.chunk(x, 2, dim=-1)

        return torch.cat(
            (
                super().forward(x0),
                -super().forward(-x1),
            ),
            dim=-1,
        )


class MonotonicMLP(MLP):
    r"""Creates a monotonic multi-layer perceptron (MonotonicMLP).

    The resulting MLP is a transformation :math:`y = f(x)` whose Jacobian entries
    :math:`\frac{\partial y_j}{\partial x_i}` are positive.

    Arguments:
        args: Positional arguments passed to :class:`MLP`.
        kwargs: Keyword arguments passed to :class:`MLP`.

    Example:
        >>> net = MonotonicMLP(3, 4, [16, 32])
        >>> net
        MonotonicMLP(
          (0): MonotonicLinear(in_features=3, out_features=16, bias=True)
          (1): TwoWayELU(alpha=1.0)
          (2): MonotonicLinear(in_features=16, out_features=32, bias=True)
          (3): TwoWayELU(alpha=1.0)
          (4): MonotonicLinear(in_features=32, out_features=4, bias=True)
        )
        >>> x = torch.randn(3)
        >>> torch.autograd.functional.jacobian(net, x)
        tensor([[1.0492, 1.3094, 1.1711],
                [1.1201, 1.3825, 1.2711],
                [0.9397, 1.1915, 1.0787],
                [1.1049, 1.3635, 1.2592]])
    """

    def __init__(self, *args, **kwargs):
        kwargs["activation"] = TwoWayELU
        kwargs["normalize"] = False

        super().__init__(*args, **kwargs)

        for layer in self:
            if isinstance(layer, Linear):
                layer.__class__ = MonotonicLinear



r"""Tests for the zuko.bayesian module."""

import math
import pytest
import torch

from pathlib import Path
from torch import randn
from typing import Sequence

from zuko.bayesian import BayesianModel
from zuko.flows import *
from zuko.nn import *


@pytest.mark.parametrize("local_trick", [False, True])
@pytest.mark.parametrize("batch", [(), (256,)])
def test_bayesian_MLP(local_trick: bool, batch: Sequence[int]):
    net = MLP(3, 5)
    bnet = BayesianModel(net)

    x = randn(*batch, 3)

    # reparameterize
    with bnet.reparameterize(local_trick=local_trick) as rnet:
        y1, y2 = rnet(x), rnet(x)

    with bnet.reparameterize(local_trick=local_trick) as rnet:
        y3 = rnet(x)

    assert y1.shape == (*batch, 5)
    assert y1.requires_grad
    assert torch.allclose(y1, y2)
    assert not torch.allclose(y1, y3)

    # sample_model
    snet = bnet.sample_model()
    y1, y2 = snet(x), snet(x)

    snet = bnet.sample_model()
    y3 = snet(x)

    assert y1.shape == (*batch, 5)
    assert y1.requires_grad
    assert torch.allclose(y1, y2)
    assert not torch.allclose(y1, y3)

    # kl_divergence
    kl = bnet.kl_divergence()

    assert torch.all(kl >= 0)


@pytest.mark.parametrize("local_trick", [False, True])
@pytest.mark.parametrize("batch", [(), (256,)])
def test_bayesian_MaskedMLP(local_trick: bool, batch: Sequence[int]):
    adjacency = randn(5, 3) < 0
    net = MaskedMLP(adjacency)
    bnet = BayesianModel(net)

    x = randn(*batch, 3)

    # reparameterize
    with bnet.reparameterize(local_trick=local_trick) as rnet:
        y1, y2 = rnet(x), rnet(x)

    with bnet.reparameterize(local_trick=local_trick) as rnet:
        y3 = rnet(x)

    assert y1.shape == (*batch, 5)
    assert y1.requires_grad
    assert torch.allclose(y1, y2)
    assert not torch.allclose(y1, y3)

    ## Jacobian
    with bnet.reparameterize() as rnet:
        x = randn(*batch, 3)
        J = torch.autograd.functional.jacobian(rnet, x)
        J = J.movedim(len(batch), -2)

    mask = torch.eye(math.prod(batch), dtype=bool)
    mask = mask.reshape(batch + batch)

    assert (J[mask][..., ~adjacency] == 0).all()
    assert (J[~mask] == 0).all()

    # sample_model
    snet = bnet.sample_model()
    y1, y2 = snet(x), snet(x)

    snet = bnet.sample_model()
    y3 = snet(x)

    assert y1.shape == (*batch, 5)
    assert y1.requires_grad
    assert torch.allclose(y1, y2)
    assert not torch.allclose(y1, y3)

    ## Jacobian
    snet = bnet.sample_model()
    x = randn(*batch, 3)
    J = torch.autograd.functional.jacobian(snet, x)
    J = J.movedim(len(batch), -2)

    mask = torch.eye(math.prod(batch), dtype=bool)
    mask = mask.reshape(batch + batch)

    assert (J[mask][..., ~adjacency] == 0).all()
    assert (J[~mask] == 0).all()


@pytest.mark.parametrize("F", [NICE, MAF, NSF, SOSPF, GF, BPF])
@pytest.mark.parametrize("local_trick", [True, False])
def test_bayesian_flows(tmp_path: Path, F: type, local_trick: bool):
    flow = F(3, 5)
    bflow = BayesianModel(flow)

    # Evaluation of log_prob
    x, c = randn(256, 3), randn(5)

    ## reparametrize
    bflow.zero_grad(set_to_none=True)

    with bflow.reparameterize(local_trick=local_trick) as rflow:
        log_p1 = rflow(c).log_prob(x)
        log_p2 = rflow(c).log_prob(x)

    with bflow.reparameterize(local_trick=local_trick) as rflow:
        log_p3 = rflow(c).log_prob(x)

    loss = -log_p3.mean()
    loss.backward()

    assert log_p1.shape == (256,)
    assert log_p1.requires_grad
    assert torch.allclose(log_p1, log_p2)
    assert not torch.allclose(log_p1, log_p3)

    for p in (*bflow.means.parameters(), *bflow.logvars.parameters()):
        assert p.grad is not None

    for p in bflow.base.parameters():
        assert p.grad is None

    ## sample_model
    bflow.zero_grad(set_to_none=True)

    sflow = bflow.sample_model()
    log_p1 = sflow(c).log_prob(x)
    log_p2 = sflow(c).log_prob(x)

    sflow = bflow.sample_model()
    log_p3 = sflow(c).log_prob(x)

    loss = -log_p3.mean()
    loss.backward()

    assert log_p1.shape == (256,)
    assert log_p1.requires_grad
    assert torch.allclose(log_p1, log_p2)
    assert not torch.allclose(log_p1, log_p3)

    for p in bflow.parameters():
        assert p.grad is None

    for p in sflow.parameters():
        assert p.grad is not None

    del sflow

    # Sampling
    with bflow.reparameterize(local_trick=local_trick) as rflow:
        x = rflow(c).sample((32,))

        assert x.shape == (32, 3)

    # Reparameterization trick
    bflow.zero_grad(set_to_none=True)

    if bflow.base(c).has_rsample:
        with bflow.reparameterize(local_trick=local_trick) as rflow:
            x = rflow(c).rsample()

        loss = x.square().sum().sqrt()
        loss.backward()

        for p in (*bflow.means.parameters(), *bflow.logvars.parameters()):
            assert p.grad is not None

        for p in bflow.base.parameters():
            assert p.grad is None

    # Invertibility
    x, c = randn(256, 3), randn(256, 5)

    with bflow.reparameterize(local_trick=local_trick) as rflow:
        t = rflow(c).transform
        z = t.inv(t(x))

        assert torch.allclose(x, z, atol=1e-4)

    # Saving
    torch.save(bflow, tmp_path / "flow.pth")

    # Loading
    bflow_bis = torch.load(tmp_path / "flow.pth", weights_only=False)

    x, c = randn(3), randn(5)

    with torch.random.fork_rng():
        with bflow.reparameterize(local_trick=local_trick) as rflow:
            log_p = rflow(c).log_prob(x)

    with torch.random.fork_rng():
        with bflow_bis.reparameterize(local_trick=local_trick) as rflow:
            log_p_bis = rflow(c).log_prob(x)

    assert torch.allclose(log_p, log_p_bis)

    # Printing
    assert repr(bflow)



r"""Autoregressive flows and transformations."""

__all__ = [
    "MAF",
    "MaskedAutoregressiveTransform",
]

import torch

from functools import partial
from math import ceil, prod
from torch import BoolTensor, LongTensor, Size, Tensor
from torch.distributions import Transform
from typing import Callable, Sequence

from .gaussianization import ElementWiseTransform
from ..distributions import DiagNormal
from ..lazy import Flow, LazyTransform, UnconditionalDistribution
from ..nn import MaskedMLP
from ..transforms import AutoregressiveTransform, DependentTransform, MonotonicAffineTransform
from ..utils import broadcast, unpack


class MaskedAutoregressiveTransform(LazyTransform):
    r"""Creates a lazy masked autoregressive transformation.

    See also:
        :class:`zuko.transforms.AutoregressiveTransform`

    References:
        | Masked Autoregressive Flow for Density Estimation (Papamakarios et al., 2017)
        | https://arxiv.org/abs/1705.07057

    Arguments:
        features: The number of features.
        context: The number of context features.
        passes: The number of sequential passes for the inverse transformation. If
            :py:`None`, use the number of features instead, making the transformation
            fully autoregressive. Coupling corresponds to :py:`passes=2`.
        order: A feature ordering. If :py:`None`, use :py:`range(features)` instead.
        adjacency: An adjacency matrix describing the transformation graph. If
            `adjacency` is provided, `order` is ignored and `passes` is replaced by the
            diameter of the graph. Its shape must be either `(features, features)`
            or `(features, features + context)`. If the shape includes context, the rightmost
            `context` columns define connections to the conditioning variables.
        univariate: The univariate transformation constructor.
        shapes: The shapes of the univariate transformation parameters.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MaskedMLP`.

    Example:
        >>> t = MaskedAutoregressiveTransform(3, 4)
        >>> t
        MaskedAutoregressiveTransform(
          (base): MonotonicAffineTransform()
          (order): [0, 1, 2]
          (hyper): MaskedMLP(
            (0): MaskedLinear(in_features=7, out_features=64, bias=True)
            (1): ReLU()
            (2): MaskedLinear(in_features=64, out_features=64, bias=True)
            (3): ReLU()
            (4): MaskedLinear(in_features=64, out_features=6, bias=True)
          )
        )
        >>> x = torch.randn(3)
        >>> x
        tensor([ 1.7428, -1.6483, -0.9920])
        >>> c = torch.randn(4)
        >>> y = t(c)(x)
        >>> t(c).inv(y)
        tensor([ 1.7428, -1.6483, -0.9920], grad_fn=<DivBackward0>)
    """

    def __new__(
        cls,
        features: int = None,
        context: int = 0,
        passes: int = None,
        order: LongTensor = None,
        adjacency: BoolTensor = None,
        *args,
        **kwargs,
    ) -> LazyTransform:
        if features is None or features > 1:
            return super().__new__(cls)
        else:
            return ElementWiseTransform(features, context, *args, **kwargs)

    def __init__(
        self,
        features: int,
        context: int = 0,
        passes: int = None,
        order: LongTensor = None,
        adjacency: BoolTensor = None,
        univariate: Callable[..., Transform] = MonotonicAffineTransform,
        shapes: Sequence[Size] = ((), ()),
        **kwargs,
    ):
        super().__init__()

        # Univariate transformation
        self.univariate = univariate
        self.shapes = shapes
        self.total = sum(prod(s) for s in shapes)

        # Adjacency
        self.register_buffer("order", None)

        if adjacency is None:
            if passes is None:
                passes = features

            if order is None:
                order = torch.arange(features)
            else:
                order = torch.as_tensor(order, dtype=int)

            assert order.ndim == 1, "'order' should be a vector."
            assert order.shape[0] == features, f"'order' should have {features} elements."

            self.passes = min(max(passes, 1), features)
            self.order = torch.div(order, ceil(features / self.passes), rounding_mode="floor")

            adjacency = self.order[:, None] > self.order
            adjacency_context = None
        else:
            adjacency = torch.as_tensor(adjacency, dtype=bool)

            assert adjacency.ndim == 2, "'adjacency' should be a matrix."
            assert adjacency.shape[0] == features, f"'adjacency' should have {features} rows."
            assert adjacency.shape[1] == features or adjacency.shape[1] == features + context, (
                f"'adjacency' should have {features} or {features + context} columns."
            )

            adjacency_context = adjacency[:, features:] if adjacency.shape[1] > features else None
            adjacency = adjacency[:, :features]

            assert adjacency.diag().all(), "'adjacency' should have ones on the diagonal."

            adjacency = adjacency * ~torch.eye(features, dtype=bool)

            self.passes = self._dag_diameter(adjacency)

        if context > 0:
            if adjacency_context is None:
                adjacency_context = torch.ones((features, context), dtype=bool)
            adjacency = torch.cat((adjacency, adjacency_context), dim=1)

        adjacency = torch.repeat_interleave(adjacency, repeats=self.total, dim=0)

        # Hyper network
        self.hyper = MaskedMLP(adjacency, **kwargs)

    @staticmethod
    def _dag_diameter(adjacency: BoolTensor) -> int:
        r"""Returns the diameter of a directed acyclic graph.

        If the graph contains cycles, this function raises an error.

        Credits:
            This code is adapted from :func:`networkx.topological_generations`.

        Arguments:
            adjacency: An adjacency matrix representing a directed graph.

        Returns:
            The diameter of the graph.
        """

        all_generations = []
        indegree = adjacency.sum(dim=1).tolist()
        zero_indegree = [n for n, d in enumerate(indegree) if d == 0]
        while zero_indegree:
            this_generation, zero_indegree = zero_indegree, []
            for node in this_generation:
                for child in adjacency[:, node].nonzero():
                    child = child.item()
                    indegree[child] -= 1
                    if indegree[child] == 0:
                        zero_indegree.append(child)
            all_generations.append(this_generation)

        assert all(d == 0 for d in indegree), "The graph contains cycles."

        return len(all_generations)

    def extra_repr(self) -> str:
        base = self.univariate(*map(torch.randn, self.shapes))

        if self.order is None:
            return "\n".join([
                f"(base): {base}",
                f"(passes): {self.passes}",
            ])
        else:
            order = self.order.tolist()

            if len(order) > 10:
                order = order[:5] + [...] + order[-5:]
                order = str(order).replace("Ellipsis", "...")

            return "\n".join([
                f"(base): {base}",
                f"(order): {order}",
            ])

    def meta(self, c: Tensor, x: Tensor) -> Transform:
        if c is not None:
            x = torch.cat(broadcast(x, c, ignore=1), dim=-1)

        phi = self.hyper(x)
        phi = phi.unflatten(-1, (-1, self.total))
        phi = unpack(phi, self.shapes)

        return DependentTransform(self.univariate(*phi), 1)

    def forward(self, c: Tensor = None) -> Transform:
        return AutoregressiveTransform(partial(self.meta, c), self.passes)


class MAF(Flow):
    r"""Creates a masked autoregressive flow (MAF).

    References:
        | Masked Autoregressive Flow for Density Estimation (Papamakarios et al., 2017)
        | https://arxiv.org/abs/1705.07057

    Arguments:
        features: The number of features.
        context: The number of context features.
        transforms: The number of autoregressive transformations.
        randperm: Whether features are randomly permuted between transformations or not.
            If :py:`False`, features are in ascending (descending) order for even
            (odd) transformations.
        kwargs: Keyword arguments passed to :class:`MaskedAutoregressiveTransform`.

    Example:
        >>> flow = MAF(3, 4, transforms=3)
        >>> flow
        MAF(
          (transform): LazyComposedTransform(
            (0): MaskedAutoregressiveTransform(
              (base): MonotonicAffineTransform()
              (order): [0, 1, 2]
              (hyper): MaskedMLP(
                (0): MaskedLinear(in_features=7, out_features=64, bias=True)
                (1): ReLU()
                (2): MaskedLinear(in_features=64, out_features=64, bias=True)
                (3): ReLU()
                (4): MaskedLinear(in_features=64, out_features=6, bias=True)
              )
            )
            (1): MaskedAutoregressiveTransform(
              (base): MonotonicAffineTransform()
              (order): [2, 1, 0]
              (hyper): MaskedMLP(
                (0): MaskedLinear(in_features=7, out_features=64, bias=True)
                (1): ReLU()
                (2): MaskedLinear(in_features=64, out_features=64, bias=True)
                (3): ReLU()
                (4): MaskedLinear(in_features=64, out_features=6, bias=True)
              )
            )
            (2): MaskedAutoregressiveTransform(
              (base): MonotonicAffineTransform()
              (order): [0, 1, 2]
              (hyper): MaskedMLP(
                (0): MaskedLinear(in_features=7, out_features=64, bias=True)
                (1): ReLU()
                (2): MaskedLinear(in_features=64, out_features=64, bias=True)
                (3): ReLU()
                (4): MaskedLinear(in_features=64, out_features=6, bias=True)
              )
            )
          )
          (base): UnconditionalDistribution(DiagNormal(loc: torch.Size([3]), scale: torch.Size([3])))
        )
        >>> c = torch.randn(4)
        >>> x = flow(c).sample()
        >>> x
        tensor([-0.5012, -1.6298,  0.3803])
        >>> flow(c).log_prob(x)
        tensor(-3.7514, grad_fn=<AddBackward0>)
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        transforms: int = 3,
        randperm: bool = False,
        **kwargs,
    ):
        orders = [
            torch.arange(features),
            torch.flipud(torch.arange(features)),
        ]

        transforms = [
            MaskedAutoregressiveTransform(
                features=features,
                context=context,
                order=torch.randperm(features) if randperm else orders[i % 2],
                **kwargs,
            )
            for i in range(transforms)
        ]

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(transforms, base)



r"""Mixture models."""

__all__ = [
    "GMM",
]

import torch
import torch.nn as nn

from math import prod
from torch import Tensor
from torch.distributions import (
    Distribution,
    MultivariateNormal,
)
from typing import Sequence, Tuple

from .distributions import DiagNormal, Mixture
from .lazy import LazyDistribution
from .nn import MLP
from .utils import unpack


class GMM(LazyDistribution):
    r"""Creates a Gaussian mixture model (GMM).

    .. math:: p(X | c) = \sum_{i = 1}^K w_i(c) \, \mathcal{N}(X | \mu_i(c), \Sigma_i(c))

    Wikipedia:
        https://wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model

    Arguments:
        features: The number of features.
        context: The number of context features.
        components: The number of components :math:`K` in the mixture.
        covariance_type: The type of covariance matrix parameterization to use.
            One of :py:`['full', 'diagonal', 'spherical']`.
        tied: Whether to tie the covariance parameters across components.
        epsilon: A numerical stability term.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        components: int = 2,
        covariance_type: str = "full",
        tied: bool = False,
        epsilon: float = 1e-6,
        **kwargs,
    ):
        super().__init__()

        self.components = components
        self.covariance_type = covariance_type
        self.tied = tied
        self.epsilon = epsilon

        shapes = _get_gmm_shapes(components, features, covariance_type, tied)

        self.shapes = shapes
        self.total = sum(prod(s) for s in shapes)

        if context > 0:
            self.hyper = MLP(context, self.total, **kwargs)
        else:
            self.phi = nn.ParameterList(torch.randn(*s) for s in shapes)

    def forward(self, c: Tensor = None) -> Distribution:
        if c is None:
            phi = self.phi
        else:
            phi = self.hyper(c)
            phi = unpack(phi, self.shapes)

        if self.covariance_type == "full":
            return self._forward_full(*phi)
        elif self.covariance_type == "diagonal":
            return self._forward_diagonal(*phi)
        elif self.covariance_type == "spherical":
            return self._forward_diagonal(*phi)
        else:
            raise ValueError(f"Unknown covariance type '{self.covariance_type}'.")

    def _forward_full(
        self, logits: Tensor, loc: Tensor, diag: Tensor, tril: Tensor
    ) -> Distribution:
        scale = torch.diag_embed(diag.exp() + self.epsilon)
        mask = torch.tril(torch.ones_like(scale, dtype=bool), diagonal=-1)
        scale = torch.masked_scatter(scale, mask, tril)

        return Mixture(MultivariateNormal(loc=loc, scale_tril=scale), logits)

    def _forward_diagonal(self, logits: Tensor, loc: Tensor, diag: Tensor) -> Distribution:
        scale = diag.exp() + self.epsilon

        return Mixture(DiagNormal(loc=loc, scale=scale), logits)

    @torch.no_grad()
    def initialize(self, x: Tensor, strategy: str):
        r"""Initializes the components of the model.

        Note:
            For more information on the clustering strategies see
            https://scikit-learn.org/dev/modules/mixture.html#choice-of-the-initialization-method

        Arguments:
            x: The feature samples, with shape :math:`(N, D)`.
            strategy: The clustering strategy. One of :py:`['random', 'kmeans', 'kmeans++']`.
        """

        N, _ = x.shape

        assert N > self.components, (
            f"The number of samples ({N}) should be larger than the number of components ({self.components})."
        )

        if strategy == "random":
            centers = _cluster_random(x, self.components)
        elif strategy == "kmeans":
            centers = _cluster_kmeans(x, self.components)
        elif strategy == "kmeans++":
            centers = _cluster_kmeans_pp(x, self.components)
        else:
            raise ValueError(f"Unkown clustering strategy '{strategy}'.")

        match = torch.cdist(x, centers).argmin(dim=-1)
        match = torch.nn.functional.one_hot(match, num_classes=self.components).to(dtype=x.dtype)

        probs = torch.sum(match, dim=0) / torch.sum(match)
        means = torch.sum(match[:, :, None] * x[:, None, :], dim=0) / torch.sum(
            match[:, :, None], dim=0
        )

        if self.covariance_type == "full":
            covs = _estimate_full_cov(x, match, self.tied)
        elif self.covariance_type == "diagonal":
            covs = _estimate_diagonal_cov(x, match, self.tied)
        elif self.covariance_type == "spherical":
            covs = _estimate_spherical_cov(x, match, self.tied)
        else:
            raise ValueError(f"Unkown covariance type '{self.covariance_type}'.")

        if torch.is_tensor(covs):
            params = (probs.log(), means, covs)
        else:
            params = (probs.log(), means, *covs)

        assert all(p.shape == s for p, s in zip(params, self.shapes))

        if hasattr(self, "hyper"):
            self.hyper[-1].weight.mul_(1e-2)
            self.hyper[-1].bias.copy_(torch.cat([p.flatten() for p in params]))
        else:
            for p, p_ in zip(self.phi, params):
                p.copy_(p_)


def _get_gmm_shapes(
    components: int,
    features: int,
    covariance_type: str,
    tied: bool,
) -> Sequence[int]:
    leading = 1 if tied else components

    shapes = [
        (components,),  # probabilities
        (components, features),  # mean
    ]

    if covariance_type == "full":
        shapes.extend([
            (leading, features),  # diagonal
            (leading, features * (features - 1) // 2),  # off diagonal
        ])
    elif covariance_type == "diagonal":
        shapes.extend([
            (leading, features),  # diagonal
        ])
    elif covariance_type == "spherical":
        shapes.extend([
            (leading, 1),  # diagonal
        ])
    else:
        raise ValueError(f"Unknown covariance type '{covariance_type}'.")

    return shapes


def _estimate_full_cov(x: Tensor, match: Tensor, tied: bool) -> Tuple[Tensor, Tensor]:
    _, D = x.shape
    _, K = match.shape

    covariances = []

    for k in range(K):
        covariances.append(torch.cov(x.T, aweights=match[:, k]))

    covariances = torch.stack(covariances)

    if tied:
        covariances = covariances.mean(dim=0, keepdim=True)

    lower = torch.linalg.cholesky(covariances)

    diag = torch.diagonal(lower, dim1=-2, dim2=-1)
    tril = lower[(..., *torch.tril_indices(D, D, offset=-1))]

    return diag.log(), tril


def _estimate_diagonal_cov(x: Tensor, match: Tensor, tied: bool) -> Tensor:
    _, D = x.shape
    _, K = match.shape

    diag = []

    for k in range(K):
        for d in range(D):
            diag.append(torch.cov(x[:, d], aweights=match[:, k]))

    diag = torch.stack(diag).reshape(K, D)

    if tied:
        diag = diag.mean(dim=0, keepdim=True)

    return diag.log()


def _estimate_spherical_cov(x: Tensor, match: Tensor, tied: bool) -> Tensor:
    diag = _estimate_diagonal_cov(x, match, tied)
    diag = diag.exp().mean(dim=-1, keepdim=True)

    return diag.log()


def _cluster_random(x: Tensor, components: int):
    N, _ = x.shape

    idx = torch.multinomial(torch.ones(N, device=x.device), components)
    centers = x[idx]

    return centers


def _cluster_kmeans(x: Tensor, components: int, iterations: int = 7):
    N, _ = x.shape

    centers = _cluster_kmeans_pp(x, components)

    for _ in range(iterations):
        match = torch.cdist(x, centers).argmin(dim=-1)
        match = torch.nn.functional.one_hot(match, num_classes=components).to(dtype=x.dtype)

        idx = torch.multinomial(torch.ones(N, device=x.device), components)
        centers = torch.where(
            torch.sum(match[:, :, None], dim=0) > 0,
            torch.sum(match[:, :, None] * x[:, None, :], dim=0)
            / torch.sum(match[:, :, None], dim=0),
            x[idx],
        )

    return centers


def _cluster_kmeans_pp(x: Tensor, components: int):
    N, _ = x.shape

    idx = torch.multinomial(torch.ones(N, device=x.device), components)
    centers = x[idx]

    mask = torch.zeros((N, components), dtype=torch.bool, device=x.device)
    mask[idx, 0] = True

    for k in range(1, components):
        dist = torch.cdist(x, centers[:k])
        dist[mask[:, :k]] = 0
        dist = dist.min(dim=-1).values

        idx = torch.multinomial(dist.square(), 1)
        centers[k] = x[idx]
        mask[idx, k] = True

    return centers



r"""Parameterizable probability distributions."""

__all__ = [
    "BoxUniform",
    "DiagNormal",
    "GeneralizedNormal",
    "Joint",
    "Maximum",
    "Minimum",
    "Mixture",
    "NormalizingFlow",
    "Sort",
    "TopK",
    "TransformedUniform",
    "Truncated",
]

import math
import torch

from textwrap import indent
from torch import Size, Tensor
from torch.distributions import *  # noqa: F403
from torch.distributions import (
    Categorical,
    Distribution,
    Independent,
    Normal,
    Transform,
    Uniform,
    constraints,
)
from torch.distributions.utils import _sum_rightmost
from typing import Tuple

Distribution._validate_args = False
Distribution.arg_constraints = {}


class NormalizingFlow(Distribution):
    r"""Creates a normalizing flow for a random variable :math:`X` towards a base
    distribution :math:`p(Z)` through a transformation :math:`f`.

    The density of a realization :math:`x` is given by the change of variables

    .. math:: p(X = x) = p(Z = f(x)) \left| \det \frac{\partial f(x)}{\partial x} \right| .

    To sample from :math:`p(X)`, realizations :math:`z \sim p(Z)` are mapped through
    the inverse transformation :math:`g = f^{-1}`.

    References:
        | A Family of Non-parametric Density Estimation Algorithms (Tabak et al., 2013)
        | https://onlinelibrary.wiley.com/doi/abs/10.1002/cpa.21423

        | Variational Inference with Normalizing Flows (Rezende et al., 2015)
        | https://arxiv.org/abs/1505.05770

        | Normalizing Flows for Probabilistic Modeling and Inference (Papamakarios et al., 2021)
        | https://arxiv.org/abs/1912.02762

    Arguments:
        transform: A transformation :math:`f`.
        base: A base distribution :math:`p(Z)`.

    Example:
        >>> d = NormalizingFlow(ExpTransform(), Gamma(2.0, 1.0))
        >>> d.sample()
        tensor(1.5157)
    """

    has_rsample = True

    def __init__(
        self,
        transform: Transform,
        base: Distribution,
    ):
        super().__init__()

        reinterpreted = transform.codomain.event_dim - len(base.event_shape)

        if reinterpreted > 0:
            base = Independent(base, reinterpreted)

        self.transform = transform
        self.base = base
        self.reinterpreted = max(-reinterpreted, 0)

    def __repr__(self) -> str:
        lines = [
            f"(transform): {self.transform}",
            f"(base): {self.base}",
        ]
        lines = indent("\n".join(lines), "  ")

        return self.__class__.__name__ + "(\n" + lines + "\n)"

    @property
    def batch_shape(self) -> Size:
        return self.base.batch_shape

    @property
    def event_shape(self) -> Size:
        return self.transform.inverse_shape(self.base.event_shape)

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(NormalizingFlow, new)
        new.transform = self.transform
        new.base = self.base.expand(batch_shape)
        new.reinterpreted = self.reinterpreted

        Distribution.__init__(new, validate_args=False)

        return new

    def log_prob(self, x: Tensor) -> Tensor:
        z, ladj = self.transform.call_and_ladj(x)
        ladj = _sum_rightmost(ladj, self.reinterpreted)

        return self.base.log_prob(z) + ladj

    def rsample(self, shape: Size = ()) -> Tensor:
        if self.base.has_rsample:
            z = self.base.rsample(shape)
        else:
            z = self.base.sample(shape)

        return self.transform.inv(z)

    def rsample_and_log_prob(self, shape: Size = ()) -> Tuple[Tensor, Tensor]:
        if self.base.has_rsample:
            z = self.base.rsample(shape)
        else:
            z = self.base.sample(shape)

        x, ladj = self.transform.inv.call_and_ladj(z)
        ladj = _sum_rightmost(ladj, self.reinterpreted)

        return x, self.base.log_prob(z) - ladj


class Joint(Distribution):
    r"""Creates a distribution for a multivariate random variable :math:`X` which
    is the concatenation of :math:`n` independent random variables :math:`Z_i`.

    .. math:: p(X = x) = \prod_i p(Z_i = x_i)

    Arguments:
        marginals: A list of distributions :math:`p(Z_i)`.

    Example:
        >>> d = Joint(Uniform(0.0, 1.0), Normal(0.0, 1.0))
        >>> d.event_shape
        torch.Size([2])
        >>> d.sample()
        tensor([0.4963, 0.2072])
    """

    has_rsample = True

    def __init__(self, *marginals: Distribution):
        super().__init__()

        batch_shape = torch.broadcast_shapes(*(m.batch_shape for m in marginals))

        self.marginals = [m.expand(batch_shape) for m in marginals]

    def __repr__(self) -> str:
        lines = map(repr, self.marginals)
        lines = indent("\n".join(lines), "  ")

        return self.__class__.__name__ + "(\n" + lines + "\n)"

    @property
    def batch_shape(self) -> Size:
        return self.marginals[0].batch_shape

    @property
    def event_shape(self) -> Size:
        return Size([sum(m.event_shape.numel() for m in self.marginals)])

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Joint, new)
        new.marginals = [m.expand(batch_shape) for m in self.marginals]

        Distribution.__init__(new, validate_args=False)

        return new

    def log_prob(self, x: Tensor) -> Tensor:
        shape = x.shape[:-1]
        i, lp = 0, 0

        for m in self.marginals:
            j = i + m.event_shape.numel()
            z = x[..., i:j].reshape(shape + m.event_shape)
            lp = lp + m.log_prob(z)
            i = j

        return lp

    def rsample(self, shape: Size = ()) -> Tensor:
        x = []

        for m in self.marginals:
            if m.has_rsample:
                z = m.rsample(shape)
            else:
                z = m.sample(shape)

            z = z.reshape(shape + m.batch_shape + (-1,))
            x.append(z)

        return torch.cat(x, dim=-1)


class Mixture(Distribution):
    r"""Creates a mixture of distributions for a random variable :math:`X`.

    .. math:: p(X = x) = \frac{1}{\sum_i w_i} \sum_i w_i \, p(Z_i = x)

    Wikipedia:
        https://wikipedia.org/wiki/Mixture_model

    Arguments:
        base: A batch of base distributions :math:`p(Z_i)`.
        logits: The unnormalized log-weights :math:`\log w_i`.

    Example:
        >>> d = Mixture(Normal(torch.randn(2), torch.ones(2)), torch.randn(2))
        >>> d.event_shape
        torch.Size([])
        >>> d.sample()
        tensor(-1.6920)
    """

    has_rsample = False

    def __init__(self, base: Distribution, logits: Tensor):
        super().__init__()

        assert base.batch_shape[-1] == logits.shape[-1]
        assert len(base.batch_shape) >= len(logits.shape)

        self.base = base
        self.logits = logits

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base})"

    @property
    def batch_shape(self) -> Size:
        return self.base.batch_shape[:-1]

    @property
    def event_shape(self) -> Size:
        return self.base.event_shape

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Mixture, new)
        new.base = self.base.expand(batch_shape + self.base.batch_shape[-1:])
        new.logits = self.logits

        Distribution.__init__(new, validate_args=False)

        return new

    def log_prob(self, x: Tensor) -> Tensor:
        x = x.unsqueeze(dim=-(len(self.event_shape) + 1))

        log_w = torch.log_softmax(self.logits, dim=-1)
        log_p = self.base.log_prob(x)

        return torch.logsumexp(log_w + log_p, dim=-1)

    def sample(self, shape: Size = ()) -> Tensor:
        with torch.no_grad():
            x = self.base.sample(shape)
            i = Categorical(logits=self.logits).expand(self.batch_shape).sample(shape)

            if i.dim() > 0:
                x = x.flatten(0, i.dim() - 1)
                x = x[torch.arange(i.numel(), device=i.device), i.flatten()]

                return x.reshape(shape + self.batch_shape + self.event_shape)
            else:
                return x[i]


class GeneralizedNormal(Distribution):
    r"""Creates a generalized normal distribution.

    .. math:: p(X = x) = \frac{\beta}{2 \Gamma(1 / \beta)} \exp(-|x|^\beta)

    Wikipedia:
        https://wikipedia.org/wiki/Generalized_normal_distribution

    Arguments:
        beta: The shape parameter :math:`\beta`.

    Example:
        >>> d = GeneralizedNormal(2.0)
        >>> d.sample()
        tensor(-0.0281)
    """

    arg_constraints = {"beta": constraints.positive}
    support = constraints.real
    has_rsample = True

    def __init__(self, beta: Tensor):
        super().__init__()

        self.beta = torch.as_tensor(beta)

    @property
    def batch_shape(self) -> Size:
        return self.beta.shape

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(GeneralizedNormal, new)
        new.beta = self.beta.expand(batch_shape)

        Distribution.__init__(new, validate_args=False)

        return new

    def log_prob(self, x: Tensor) -> Tensor:
        return torch.log(self.beta / 2) - torch.lgamma(1 / self.beta) - abs(x) ** self.beta

    def rsample(self, shape: Size = ()) -> Tensor:
        beta = self.beta.expand(shape + self.beta.shape)
        x = torch._standard_gamma(1 / beta) ** (1 / beta)
        x = x * torch.sign(2 * torch.rand_like(x) - 1)
        return x


class DiagNormal(Independent):
    r"""Creates a multivariate normal distribution parametrized by the variables
    mean :math:`\mu` and standard deviation :math:`\sigma`, but assumes no
    correlation between the variables.

    Arguments:
        loc: The mean :math:`\mu` of the variables.
        scale: The standard deviation :math:`\sigma` of the variables.
        ndims: The number of batch dimensions to interpret as event dimensions.

    Example:
        >>> d = DiagNormal(torch.zeros(3), torch.ones(3))
        >>> d.event_shape
        torch.Size([3])
        >>> d.sample()
        tensor([ 1.5410, -0.2934, -2.1788])
    """

    def __init__(self, loc: Tensor, scale: Tensor, ndims: int = 1):
        super().__init__(Normal(torch.as_tensor(loc), torch.as_tensor(scale)), ndims)

    def __repr__(self) -> str:
        return "Diag" + repr(self.base_dist)

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(DiagNormal, new)
        return super().expand(batch_shape, new)


class BoxUniform(Independent):
    r"""Creates a distribution for a multivariate random variable :math:`X`
    distributed uniformly over an hypercube domain. Formally,

    .. math:: l_i \leq X_i < u_i ,

    where :math:`l_i` and :math:`u_i` are respectively the lower and upper bounds
    of the domain in the :math:`i`-th dimension.

    Arguments:
        lower: The lower bounds (inclusive).
        upper: The upper bounds (exclusive).
        ndims: The number of batch dimensions to interpret as event dimensions.

    Example:
        >>> d = BoxUniform(-torch.ones(3), torch.ones(3))
        >>> d.event_shape
        torch.Size([3])
        >>> d.sample()
        tensor([-0.0075,  0.5364, -0.8230])
    """

    def __init__(self, lower: Tensor, upper: Tensor, ndims: int = 1):
        super().__init__(Uniform(torch.as_tensor(lower), torch.as_tensor(upper)), ndims)

    def __repr__(self) -> str:
        return "Box" + repr(self.base_dist)

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(BoxUniform, new)
        return super().expand(batch_shape, new)


class TransformedUniform(NormalizingFlow):
    r"""Creates a distribution for a random variable :math:`X`, whose
    transformation :math:`f(X)` is uniformly distributed over the interval
    :math:`[f(l), f(u)]`.

    .. math:: p(X = x) = \frac{1}{f(u) - f(l)}
        \begin{cases}
            f'(x) & \text{if } f(l) \leq f(x) < f(u) \\
            0 & \text{otherwise}
        \end{cases}

    Arguments:
        f: A transformation :math:`f`, monotonically increasing over :math:`[l, u]`.
        lower: A lower bound :math:`l` (inclusive).
        upper: An upper bound :math:`u` (exclusive).

    Example:
        >>> d = TransformedUniform(ExpTransform(), -1.0, 1.0)
        >>> d.sample()
        tensor(0.4281)
    """

    def __init__(self, f: Transform, lower: Tensor, upper: Tensor):
        super().__init__(f, Uniform(*map(f, map(torch.as_tensor, (lower, upper)))))

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(TransformedUniform, new)
        return super().expand(batch_shape, new)


class Truncated(Distribution):
    r"""Truncates a base distribution :math:`p(X)` between a lower bound
    :math:`l` and an upper bound :math:`u`.

    .. math:: p(X = x | l \leq X < u) = \frac{1}{P(X \leq u) - P(X \leq l)}
        \begin{cases}
            p(X = x) & \text{if } l \leq x < u \\
            0 & \text{otherwise}
        \end{cases}

    Arguments:
        base: A base distribution :math:`p(X)`.
        lower: A lower bound :math:`l` (inclusive).
        upper: An upper bound :math:`u` (exclusive).

    Example:
        >>> d = Truncated(Normal(0.0, 1.0), 1.0, 2.0)
        >>> d.sample()
        tensor(1.3333)
    """

    has_rsample = True

    def __init__(
        self,
        base: Distribution,
        lower: Tensor = float("-inf"),
        upper: Tensor = float("+inf"),
    ):
        super().__init__()

        assert not base.event_shape, "'base' has to be univariate"

        self.base = base
        self.uniform = Uniform(base.cdf(lower), base.cdf(upper))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base})"

    @property
    def batch_shape(self) -> Size:
        return self.base.batch_shape

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Truncated, new)
        new.base = self.base.expand(batch_shape)
        new.uniform = self.uniform.expand(batch_shape)

        Distribution.__init__(new, validate_args=False)

        return new

    def cdf(self, x: Tensor) -> Tensor:
        return self.uniform.cdf(self.base.cdf(x))

    def log_prob(self, x: Tensor) -> Tensor:
        return self.uniform.log_prob(self.base.cdf(x)) + self.base.log_prob(x)

    def rsample(self, shape: Size = ()) -> Tensor:
        return self.base.icdf((1 - 2e-6) * self.uniform.rsample(shape) + 1e-6)


class Sort(Distribution):
    r"""Creates a distribution for a :math:`n`-d random variable :math:`X`, whose elements
    :math:`X_i` are :math:`n` draws from a base distribution :math:`p(Z)`, ordered
    such that :math:`X_i \leq X_{i + 1}`.

    .. math:: p(X = x) = \begin{cases}
            n! \, \prod_{i = 1}^n p(Z = x_i) & \text{if $x$ is ordered} \\
            0 & \text{otherwise}
        \end{cases}

    Arguments:
        base: A base distribution :math:`p(Z)`.
        n: The number of draws :math:`n`.
        descending: Whether the elements are sorted in descending order or not.

    Example:
        >>> d = Sort(Normal(0.0, 1.0), 3)
        >>> d.event_shape
        torch.Size([3])
        >>> d.sample()
        tensor([-2.1788, -0.2934,  1.5410])
    """

    def __init__(
        self,
        base: Distribution,
        n: int = 2,
        descending: bool = False,
    ):
        super().__init__()

        assert not base.event_shape, "'base' has to be univariate"

        self.base = base
        self.n = n
        self.descending = descending
        self.log_fact = math.log(math.factorial(n))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base}, {self.n})"

    @property
    def batch_shape(self) -> Size:
        return self.base.batch_shape

    @property
    def event_shape(self) -> Size:
        return Size([self.n])

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Sort, new)
        new.base = self.base.expand(batch_shape)
        new.n = self.n
        new.descending = self.descending
        new.log_fact = self.log_fact

        Distribution.__init__(new, validate_args=False)

        return new

    def log_prob(self, x: Tensor) -> Tensor:
        x = torch.movedim(x, -1, 0)

        if self.descending:
            ordered = x[:-1] >= x[1:]
        else:
            ordered = x[:-1] <= x[1:]

        ordered = ordered.all(dim=0)

        return ordered.log() + self.log_fact + self.base.log_prob(x).sum(dim=0)

    def sample(self, shape: Size = ()) -> Tensor:
        x = torch.movedim(self.base.sample((self.n, *shape)), 0, -1)
        x = torch.sort(x, dim=-1, descending=self.descending).values

        return x


class TopK(Sort):
    r"""Creates a distribution for a :math:`k`-d random variable :math:`X`, whose elements
    :math:`X_i` are the top :math:`k` among :math:`n` draws from a base distribution
    :math:`p(Z)`, ordered such that :math:`X_i \leq X_{i + 1}`.

    .. math:: p(X = x) = \begin{cases}
            \frac{n!}{(n - k)!} \, \prod_{i = 1}^k p(Z = x_i)
                \, P(Z \geq x_k)^{n - k} & \text{if $x$ is ordered} \\
            0 & \text{otherwise}
        \end{cases}

    Arguments:
        base: A base distribution :math:`p(Z)`.
        k: The number of selected elements :math:`k`.
        n: The number of draws :math:`n`.
        kwargs: Keyword arguments passed to :class:`Sort`.

    Example:
        >>> d = TopK(Normal(0.0, 1.0), 2, 3)
        >>> d.event_shape
        torch.Size([2])
        >>> d.sample()
        tensor([-2.1788, -0.2934])
    """

    def __init__(
        self,
        base: Distribution,
        k: int = 1,
        n: int = 2,
        **kwargs,
    ):
        super().__init__(base, n, **kwargs)

        assert 1 <= k < n, "k has to be in [1, n)"

        self.k = k
        self.log_fact = self.log_fact - math.log(math.factorial(n - k))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base}, {self.k}, {self.n})"

    @property
    def event_shape(self) -> Size:
        return Size([self.k])

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(TopK, new)
        new.k = self.k
        return super().expand(batch_shape, new)

    def log_prob(self, x: Tensor) -> Tensor:
        cdf = self.base.cdf(x[..., -1])

        if not self.descending:
            cdf = 1 - cdf

        return (self.n - self.k) * cdf.log() + super().log_prob(x)

    def sample(self, shape: Size = ()) -> Tensor:
        return super().sample(shape)[..., : self.k]


class Minimum(TopK):
    r"""Creates a distribution for a random variable :math:`X`, which is the
    minimum among :math:`n` draws from a base distribution :math:`p(Z)`.

    .. math:: p(X = x) = n \, p(Z = x) \, P(Z \geq x)^{n - 1}

    Arguments:
        base: A base distribution :math:`p(Z)`.
        n: The number of draws :math:`n`.

    Example:
        >>> d = Minimum(Normal(0.0, 1.0), 3)
        >>> d.event_shape
        torch.Size([])
        >>> d.sample()
        tensor(-2.1788)
    """

    def __init__(self, base: Distribution, n: int = 2):
        super().__init__(base, 1, n)

        self.descending = False

    def __repr__(self) -> str:
        return Sort.__repr__(self)

    @property
    def event_shape(self) -> Size:
        return Size([])

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Minimum, new)
        return super().expand(batch_shape, new)

    def log_prob(self, x: Tensor) -> Tensor:
        return super().log_prob(x.unsqueeze(dim=-1))

    def sample(self, shape: Size = ()) -> Tensor:
        return super().sample(shape).squeeze(dim=-1)


class Maximum(Minimum):
    r"""Creates a distribution for a random variable :math:`X`, which is the
    maximum among :math:`n` draws from a base distribution :math:`p(Z)`.

    .. math:: p(X = x) = n \, p(Z = x) \, P(Z \leq x)^{n - 1}

    Arguments:
        base: A base distribution :math:`p(Z)`.
        n: The number of draws :math:`n`.

    Example:
        >>> d = Maximum(Normal(0.0, 1.0), 3)
        >>> d.event_shape
        torch.Size([])
        >>> d.sample()
        tensor(1.5410)
    """

    def __init__(self, base: Distribution, n: int = 2):
        super().__init__(base, n)

        self.descending = True

    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
        new = self._get_checked_instance(Maximum, new)
        return super().expand(batch_shape, new)



r"""Gaussianization flows."""

__all__ = [
    "GF",
    "ElementWiseTransform",
]

import torch
import torch.nn as nn

from math import prod
from torch import Size, Tensor
from torch.distributions import Transform
from typing import Callable, Sequence

from ..distributions import DiagNormal
from ..lazy import Flow, LazyTransform, UnconditionalDistribution, UnconditionalTransform
from ..nn import MLP
from ..transforms import (
    DependentTransform,
    GaussianizationTransform,
    MonotonicAffineTransform,
    RotationTransform,
)
from ..utils import unpack


class ElementWiseTransform(LazyTransform):
    r"""Creates a lazy element-wise transformation.

    Arguments:
        features: The number of features.
        context: The number of context features.
        univariate: The univariate transformation constructor.
        shapes: The shapes of the univariate transformation parameters.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.

    Example:
        >>> t = ElementWiseTransform(3, 4)
        >>> t
        ElementWiseTransform(
          (base): MonotonicAffineTransform()
          (hyper): MLP(
            (0): Linear(in_features=4, out_features=64, bias=True)
            (1): ReLU()
            (2): Linear(in_features=64, out_features=64, bias=True)
            (3): ReLU()
            (4): Linear(in_features=64, out_features=6, bias=True)
          )
        )
        >>> x = torch.randn(3)
        >>> x
        tensor([ 0.0303,  0.3644, -1.1831])
        >>> c = torch.randn(4)
        >>> y = t(c)(x)
        >>> t(c).inv(y)
        tensor([ 0.0303,  0.3644, -1.1831], grad_fn=<DivBackward0>)
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        univariate: Callable[..., Transform] = MonotonicAffineTransform,
        shapes: Sequence[Size] = ((), ()),
        **kwargs,
    ):
        super().__init__()

        self.univariate = univariate
        self.shapes = shapes
        self.total = sum(prod(s) for s in shapes)

        if context > 0:
            self.hyper = MLP(context, features * self.total, **kwargs)
        else:
            self.phi = nn.ParameterList(torch.randn(features, *s) for s in shapes)

    def extra_repr(self) -> str:
        base = self.univariate(*map(torch.randn, self.shapes))

        return "\n".join([
            f"(base): {base}",
        ])

    def forward(self, c: Tensor = None) -> Transform:
        if c is None:
            phi = self.phi
        else:
            phi = self.hyper(c)
            phi = phi.unflatten(-1, (-1, self.total))
            phi = unpack(phi, self.shapes)

        return DependentTransform(self.univariate(*phi), 1)


class GF(Flow):
    r"""Creates a gaussianization flow (GF).

    Warning:
        Invertibility is only guaranteed for features within the interval :math:`[-10,
        10]`. It is recommended to standardize features (zero mean, unit variance)
        before training.

    See also:
        :class:`zuko.transforms.GaussianizationTransform`

    References:
        | Gaussianization Flows (Meng et al., 2020)
        | https://arxiv.org/abs/2003.01941

    Arguments:
        features: The number of features.
        context: The number of context features.
        transforms: The number of coupling transformations.
        components: The number of mixture components in each transformation.
        kwargs: Keyword arguments passed to :class:`ElementWiseTransform`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        transforms: int = 3,
        components: int = 8,
        **kwargs,
    ):
        transforms = [
            ElementWiseTransform(
                features=features,
                context=context,
                univariate=GaussianizationTransform,
                shapes=[(components,), (components,)],
                **kwargs,
            )
            for _ in range(transforms)
        ]

        for i in reversed(range(1, len(transforms))):
            transforms.insert(
                i,
                UnconditionalTransform(
                    RotationTransform,
                    torch.randn(features, features),
                ),
            )

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(transforms, base)



r"""Tests for the zuko.distributions module."""

import torch

from torch.distributions import *

from zuko.distributions import *


def test_distributions():
    ds = [
        NormalizingFlow(ExpTransform(), Gamma(2.0, 1.0)),
        Joint(Uniform(0.0, 1.0), Normal(0.0, 1.0)),
        Mixture(Normal(torch.randn(2), torch.ones(2)), torch.randn(2)),
        GeneralizedNormal(2.0),
        DiagNormal(torch.zeros(2), torch.ones(2)),
        BoxUniform(-torch.ones(2), torch.ones(2)),
        TransformedUniform(ExpTransform(), -1.0, 1.0),
        Truncated(Normal(0.0, 1.0), 1.0, 2.0),
        Sort(Normal(0.0, 1.0), 2),
        TopK(Normal(0.0, 1.0), 2, 3),
        Minimum(Normal(0.0, 1.0), 3),
        Maximum(Normal(0.0, 1.0), 3),
    ]

    shape = (2**18,)

    for d in ds:
        assert d.batch_shape == (), d

        # Shapes
        x = d.sample(shape)

        assert x.shape == shape + d.event_shape, d

        log_p = d.log_prob(x)

        assert log_p.shape == shape, d

        # Expectation
        lower, upper = x.min(dim=0).values, x.max(dim=0).values
        width = upper - lower

        x = Uniform(lower - width / 2, upper + width / 2).sample(shape)

        p = d.log_prob(x).exp().mean() * (2 * width).prod()

        assert (0.9 <= p) and (p <= 1.1), d

        # Expand
        d = d.expand((32,))

        assert d.batch_shape == (32,), d

        x = d.sample()

        assert x.shape == d.batch_shape + d.event_shape, d

        log_p = d.log_prob(x)

        assert log_p.shape == d.batch_shape, d



r"""Lazy distributions and transformations."""

from __future__ import annotations

__all__ = [
    "Flow",
    "LazyComposedTransform",
    "LazyDistribution",
    "LazyTransform",
    "Unconditional",
    "UnconditionalDistribution",
    "UnconditionalTransform",
]

import abc
import torch.nn as nn
import warnings

from torch import Tensor
from torch.distributions import Distribution, Transform
from typing import Any, Callable, Sequence, Union

from .distributions import NormalizingFlow
from .transforms import ComposedTransform
from .utils import Partial


class LazyDistribution(nn.Module, abc.ABC):
    r"""Abstract lazy distribution.

    A lazy distribution is a module that builds and returns a distribution
    :math:`p(X | c)` within its forward pass, given a context :math:`c`.

    See also:
        :class:`torch.distributions.distribution.Distribution`
    """

    @abc.abstractmethod
    def forward(self, c: Any = None) -> Distribution:
        r"""
        Arguments:
            c: A context :math:`c`.

        Returns:
            A distribution :math:`p(X | c)`.
        """

        pass


class LazyTransform(nn.Module, abc.ABC):
    r"""Abstract lazy transformation.

    A lazy transformation is a module that builds and returns a transformation
    :math:`y = f(x | c)` within its forward pass, given a context :math:`c`.

    See also:
        :class:`torch.distributions.transforms.Transform`
    """

    @abc.abstractmethod
    def forward(self, c: Any = None) -> Transform:
        r"""
        Arguments:
            c: A context :math:`c`.

        Returns:
            A transformation :math:`y = f(x | c)`.
        """

        pass

    @property
    def inv(self) -> LazyTransform:
        r"""A lazy inverse transformation :math:`x = f^{-1}(y | c)`."""

        return LazyInverse(self)


class LazyInverse(LazyTransform):
    r"""Creates a lazy inverse transformation.

    Arguments:
        transform: A lazy transformation :math:`y = f(x | c)`.
    """

    def __init__(self, transform: LazyTransform):
        super().__init__()

        self.transform = transform

    def forward(self, c: Any = None) -> Transform:
        return self.transform(c).inv

    @property
    def inv(self) -> LazyTransform:
        return self.transform


class LazyComposedTransform(LazyTransform):
    r"""Creates a lazy composed transformation.

    See also:
        :class:`zuko.transforms.ComposedTransform`

    Arguments:
        transforms: A sequence of lazy transformations :math:`f_i`.
    """

    def __init__(self, *transforms: LazyTransform):
        super().__init__()

        self.transforms = nn.ModuleList(transforms)

    def __repr__(self) -> str:
        return repr(self.transforms).replace("ModuleList", "LazyComposedTransform", 1)

    def forward(self, c: Any = None) -> Transform:
        r"""
        Arguments:
            c: A context :math:`c`.

        Returns:
            A transformation :math:`y = f_n \circ \dots \circ f_0(x | c)`.
        """

        return ComposedTransform(*(t(c) for t in self.transforms))


class Flow(LazyDistribution):
    r"""Creates a lazy normalizing flow.

    See also:
        :class:`zuko.distributions.NormalizingFlow`

    Arguments:
        transform: A lazy transformation or sequence of lazy transformations.
        base: A lazy distribution.
    """

    def __init__(
        self,
        transform: Union[LazyTransform, Sequence[LazyTransform]],
        base: LazyDistribution,
    ):
        super().__init__()

        if isinstance(transform, LazyTransform):
            self.transform = transform
        else:
            self.transform = LazyComposedTransform(*transform)

        self.base = base

    def forward(self, c: Tensor = None) -> NormalizingFlow:
        r"""
        Arguments:
            c: A context :math:`c`.

        Returns:
            A normalizing flow :math:`p(X | c)`.
        """

        transform = self.transform(c)

        if c is None:
            base = self.base(c)
        else:
            base = self.base(c).expand(c.shape[:-1])

        return NormalizingFlow(transform, base)


class Unconditional(nn.Module):
    r"""Creates an unconditional lazy module from a constructor.

    Typically, the constructor returns a distribution or transformation. The positional
    arguments of the constructor are registered as buffers or parameters.

    Warning:
        :class:`Unconditional` is deprecated and will be removed in the future. Use
        :class:`UnconditionalDistribution` or  :class:`UnconditionalTransform` instead.

    Arguments:
        meta: An arbitrary constructor function.
        args: The positional tensor arguments passed to `meta`.
        buffer: Whether tensors are registered as buffers or parameters.
        kwargs: The keyword arguments passed to `meta`.
    """

    def __init__(
        self,
        meta: Callable[..., Any],
        *args: Tensor,
        buffer: bool = False,
        **kwargs,
    ):
        super().__init__()

        warnings.warn(
            (
                "'Unconditional' is deprecated and will be removed in the future. "
                "Use 'UnconditionalDistribution' or 'UnconditionalTransform' instead."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

        self.meta = meta

        for i, arg in enumerate(args):
            if buffer:
                self.register_buffer(f"_{i}", arg)
            else:
                self.register_parameter(f"_{i}", nn.Parameter(arg))

        self.kwargs = kwargs

    def extra_repr(self) -> str:
        if isinstance(self.meta, nn.Module):
            return ""
        else:
            return repr(self.forward())

    def forward(self, c: Tensor = None) -> Any:
        r"""
        Arguments:
            c: A context :math:`c`. This argument is always ignored.

        Returns:
            :py:`meta(*args, **kwargs)`
        """

        return self.meta(
            *self._parameters.values(),
            *self._buffers.values(),
            **self.kwargs,
        )


class UnconditionalDistribution(Partial, LazyDistribution):
    r"""Creates an unconditional lazy distribution from a constructor.

    The arguments of the constructor are registered as buffers or parameters.

    Arguments:
        f: A distribution constructor. If `f` is a module, it is registered as a submodule.
        args: The positional arguments passed to `f`.
        buffer: Whether tensor arguments are registered as buffers or parameters.
        kwargs: The keyword arguments passed to `f`.

    Examples:
        >>> f = zuko.distributions.DiagNormal
        >>> mu, sigma = torch.zeros(3), torch.ones(3)
        >>> base = UnconditionalDistribution(f, mu, sigma, buffer=True)
        >>> base()
        DiagNormal(loc: torch.Size([3]), scale: torch.Size([3]))
        >>> base().sample()
        tensor([ 1.5410, -0.2934, -2.1788])
    """

    def __init__(
        self,
        f: Callable[..., Distribution],
        *args: Any,
        buffer: bool = False,
        **kwargs: Any,
    ):
        super().__init__(f, *args, buffer=buffer, **kwargs)

    def extra_repr(self) -> str:
        if isinstance(self.f, nn.Module):
            return ""
        else:
            return repr(self.forward())

    def forward(self, c: Tensor = None) -> Distribution:
        r"""
        Arguments:
            c: A context :math:`c`. This argument is always ignored.

        Returns:
            :py:`self.f(*self.args, **self.kwargs)`
        """

        return super().forward()


class UnconditionalTransform(Partial, LazyTransform):
    r"""Creates an unconditional lazy transformation from a constructor.

    The arguments of the constructor are registered as buffers or parameters.

    Arguments:
        f: A transformation constructor. If `f` is a module, it is registered as a submodule.
        args: The positional arguments passed to `f`.
        buffer: Whether tensor arguments are registered as buffers or parameters.
        kwargs: The keyword arguments passed to `f`.

    Examples:
        >>> f = zuko.transforms.ExpTransform
        >>> t = UnconditionalTransform(f)
        >>> t()
        ExpTransform()
        >>> x = torch.randn(3)
        >>> t()(x)
        tensor([4.6692, 0.7457, 0.1132])
    """

    def __init__(
        self,
        f: Callable[..., Transform],
        *args: Any,
        buffer: bool = False,
        **kwargs: Any,
    ):
        super().__init__(f, *args, buffer=buffer, **kwargs)

    def extra_repr(self) -> str:
        if isinstance(self.f, nn.Module):
            return ""
        else:
            return repr(self.forward())

    def forward(self, c: Tensor = None) -> Transform:
        r"""
        Arguments:
            c: A context :math:`c`. This argument is always ignored.

        Returns:
            :py:`self.f(*self.args, **self.kwargs)`
        """

        return super().forward()



r"""Tests for the zuko.utils module."""

import pytest
import torch

from torch import randn

from zuko.utils import *


def test_bisection():
    alpha = torch.tensor(1.0, requires_grad=True)

    f = lambda x: torch.cos(alpha * x)
    g = lambda y: torch.acos(y) / alpha

    y = 1.98 * torch.rand(256, requires_grad=True) - 0.99
    x = bisection(f, y, torch.pi, 0.0, n=24, phi=(alpha,))

    assert x.shape == y.shape
    assert torch.allclose(x, g(y), atol=1e-6)
    assert torch.allclose(f(x), y, atol=1e-6)

    # Gradients
    grad_y, grad_alpha = torch.autograd.grad(x.sum(), (y, alpha))
    dy, dalpha = torch.autograd.grad(g(y).sum(), (y, alpha))

    assert torch.allclose(grad_y, dy, atol=1e-6)
    assert torch.allclose(grad_alpha, dalpha, atol=1e-6)


def test_broadcast():
    # Trivial
    a = randn(2, 3)
    (b,) = broadcast(a)

    assert a.shape == b.shape
    assert (a == b).all()

    # Standard
    a, b, c = randn(1).squeeze(), randn(2), randn(3, 1)
    d, e, f = broadcast(a, b, c)

    assert d.shape == e.shape == f.shape == (3, 2)
    assert (a == d).all() and (b == e).all() and (c == f).all()

    # Invalid
    with pytest.raises(RuntimeError):
        a, b = randn(2), randn(3)
        d, e = broadcast(a, b)

    # Ignore last dimension
    a, b = randn(2), randn(3, 4)
    c, d = broadcast(a, b, ignore=1)

    assert c.shape == (3, 2) and d.shape == (3, 4)
    assert (a == c).all() and (b == d).all()

    # Ignore mixed dimensions
    a, b = randn(2, 3), randn(3, 4)
    c, d = broadcast(a, b, ignore=[0, 1])

    assert c.shape == (2, 3) and d.shape == (2, 3, 4)
    assert (a == c).all() and (b == d).all()


def test_gauss_legendre():
    # Polynomial
    alpha = torch.tensor(1.0, requires_grad=True)

    f = lambda x: alpha * x**5 - x**2
    F = lambda x: alpha * x**6 / 6 - x**3 / 3

    a, b = randn(2, 256, requires_grad=True)
    area = gauss_legendre(f, a, b, n=3, phi=(alpha,))

    assert torch.allclose(area, F(b) - F(a), atol=1e-6)

    # Gradients
    grad_a, grad_b, grad_alpha = torch.autograd.grad(area.sum(), (a, b, alpha))
    da, db, dalpha = torch.autograd.grad((F(b) - F(a)).sum(), (a, b, alpha))

    assert torch.allclose(grad_a, da, atol=1e-6)
    assert torch.allclose(grad_b, db, atol=1e-6)
    assert torch.allclose(grad_alpha, dalpha, atol=1e-6)


def test_odeint():
    # Linear
    alpha = torch.tensor(1.0, requires_grad=True)
    t = torch.tensor(3.0, requires_grad=True)

    f = lambda t, x: -alpha * x
    F = lambda t, x: x * (-alpha * t).exp()

    x0 = randn(256, requires_grad=True)
    xt = odeint(f, x0, torch.zeros_like(t), t, phi=(alpha,), atol=1e-7, rtol=1e-6)

    assert xt.shape == x0.shape
    assert torch.allclose(xt, F(t, x0), atol=1e-6)

    # Gradients
    grad_x0, grad_t, grad_alpha = torch.autograd.grad(xt.sum(), (x0, t, alpha))
    dx0, dt, dalpha = torch.autograd.grad(F(t, x0).sum(), (x0, t, alpha))

    assert torch.allclose(grad_x0, dx0, atol=1e-6)
    assert torch.allclose(grad_t, dt, atol=1e-6)
    assert torch.allclose(grad_alpha, dalpha, atol=1e-6)


def test_unpack():
    # Normal
    x = randn(26)
    y, z = unpack(x, ((1, 2, 3), (4, 5)))

    assert y.shape == (1, 2, 3)
    assert z.shape == (4, 5)

    # Batched
    x = randn(7, 26)
    y, z = unpack(x, ((1, 2, 3), (4, 5)))

    assert y.shape == (7, 1, 2, 3)
    assert z.shape == (7, 4, 5)



r"""Tests for the zuko.flows module."""

import pytest
import torch

from functools import partial
from pathlib import Path
from torch import randn

from zuko.flows import *


@pytest.mark.parametrize("F", [NICE, MAF, NSF, SOSPF, NAF, UNAF, CNF, GF, BPF])
def test_flows(tmp_path: Path, F: type):
    flow = F(3, 5)

    # Evaluation of log_prob
    x, c = randn(256, 3), randn(5)
    log_p = flow(c).log_prob(x)

    assert log_p.shape == (256,)
    assert log_p.requires_grad

    flow.zero_grad(set_to_none=True)
    loss = -log_p.mean()
    loss.backward()

    for p in flow.parameters():
        assert p.grad is not None

    # Sampling
    x = flow(c).sample((32,))

    assert x.shape == (32, 3)

    # Reparameterization trick
    if flow(c).has_rsample:
        x = flow(c).rsample()

        flow.zero_grad(set_to_none=True)
        loss = x.square().sum().sqrt()
        loss.backward()

        for p in flow.parameters():
            assert p.grad is not None

    # Invertibility
    x, c = randn(256, 3), randn(256, 5)
    t = flow(c).transform
    z = t.inv(t(x))

    assert torch.allclose(x, z, atol=1e-4)

    # Saving
    torch.save(flow, tmp_path / "flow.pth")

    # Loading
    flow_bis = torch.load(tmp_path / "flow.pth", weights_only=False)

    x, c = randn(3), randn(5)

    with torch.random.fork_rng():
        log_p = flow(c).log_prob(x)

    with torch.random.fork_rng():
        log_p_bis = flow_bis(c).log_prob(x)

    assert torch.allclose(log_p, log_p_bis)

    # Printing
    assert repr(flow)


def test_triangular_transforms():
    order = torch.randperm(5)

    adjacency = torch.rand((5, 5)) < 0.25
    adjacency = adjacency + torch.eye(5, dtype=bool)
    adjacency = torch.tril(adjacency)
    adjacency[1, 0] = True
    adjacency = adjacency[order, :][:, order]

    Ts = [
        ElementWiseTransform,
        GeneralCouplingTransform,
        partial(GeneralCouplingTransform, mask=order % 2),
        MaskedAutoregressiveTransform,
        partial(MaskedAutoregressiveTransform, passes=2),
        partial(MaskedAutoregressiveTransform, order=order),
        partial(MaskedAutoregressiveTransform, adjacency=adjacency),
    ]

    for T in Ts:
        # Without context
        t = T(5)
        x = randn(64, 5)
        y = t()(x)

        assert y.shape == x.shape, T
        assert y.requires_grad, T
        assert torch.allclose(t().inv(y), x, atol=1e-4), T

        # With context
        t = T(5, 7)
        x, c = randn(64, 5), randn(7)
        y = t(c)(x)

        assert y.shape == x.shape, T
        assert y.requires_grad, T
        assert torch.allclose(t(c).inv(y), x, atol=1e-4), T

        # Jacobian
        t = T(5)
        x = randn(5)
        y = t()(x)

        J = torch.autograd.functional.jacobian(t(), x)
        ladj = torch.linalg.slogdet(J).logabsdet

        assert torch.allclose(t().log_abs_det_jacobian(x, y), ladj, atol=1e-4), T
        assert torch.allclose(J.diag().abs().log().sum(), ladj, atol=1e-4), T


def test_adjacency_matrix():
    T = MaskedAutoregressiveTransform

    # With a valid adjacency matrix
    order = torch.randperm(5)

    adjacency = torch.rand((5, 5)) < 0.25
    adjacency = adjacency + torch.eye(5, dtype=bool)
    adjacency = torch.tril(adjacency)
    adjacency[1, 0] = True
    adjacency = adjacency[order, :][:, order]

    t = T(5, adjacency=adjacency)
    x = randn(5)

    J = torch.autograd.functional.jacobian(t(), x)

    assert (J[~adjacency] == 0).all()

    # With False in the diagonal
    adjacency_invalid = adjacency.clone()
    adjacency_invalid[0, 0] = False

    with pytest.raises(AssertionError, match="'adjacency' should have ones on the diagonal."):
        t = T(5, adjacency=adjacency_invalid)

    # With cycles
    adjacency_invalid = adjacency.clone()
    adjacency_invalid[0, 1] = True
    adjacency_invalid[1, 0] = True

    with pytest.raises(AssertionError, match="The graph contains cycles."):
        t = T(5, adjacency=adjacency_invalid)


def test_context_adjacency_matrix():
    T = MaskedAutoregressiveTransform

    # With a valid adjacency matrix
    order = torch.randperm(5)
    adjacency = torch.rand((5, 5)) < 0.25
    adjacency = adjacency + torch.eye(5, dtype=bool)
    adjacency = torch.tril(adjacency)
    adjacency = adjacency[order, :][:, order]

    # context adjacency
    adjacency_context = torch.rand((5, 2)) < 0.25
    adjacency_valid = torch.cat((adjacency, adjacency_context), dim=1)

    t = T(features=5, context=2, adjacency=adjacency_valid)

    x, c = randn(5), randn(2)
    y = t(c)(x)

    assert y.shape == x.shape
    assert y.requires_grad
    assert torch.allclose(t(c).inv(y), x, atol=1e-4)

    J = torch.autograd.functional.jacobian(t(c), x)

    assert (J[~adjacency] == 0).all()

    ladj = torch.linalg.slogdet(J).logabsdet

    assert torch.allclose(t(c).log_abs_det_jacobian(x, y), ladj, atol=1e-4)
    assert torch.allclose(J.diag().abs().log().sum(), ladj, atol=1e-4)

    adjacency_invalid_context = torch.rand((5, 1)) < 0.25
    adjacency_invalid = torch.cat((adjacency, adjacency_invalid_context), dim=1)

    with pytest.raises(AssertionError, match="'adjacency' should have 5 or 7 columns."):
        t = T(features=5, context=2, adjacency=adjacency_invalid)



r"""Parameterized flows and transformations."""

from .autoregressive import MAF, MaskedAutoregressiveTransform
from .continuous import CNF, FFJTransform
from .core import (
    Flow,
    LazyDistribution,
    LazyInverse,
    LazyTransform,
    UnconditionalDistribution,
    UnconditionalTransform,
)
from .coupling import NICE, GeneralCouplingTransform, RealNVP
from .gaussianization import GF, ElementWiseTransform
from .mixture import GMM
from .neural import NAF, UNAF
from .polynomial import BPF, SOSPF
from .spline import NCSF, NSF



r"""Coupling flows and transformations."""

__all__ = [
    "NICE",
    "GeneralCouplingTransform",
    "RealNVP",
]

import torch

from functools import partial
from math import prod
from torch import BoolTensor, Size, Tensor
from torch.distributions import Transform
from typing import Callable, Sequence

from .gaussianization import ElementWiseTransform
from ..distributions import DiagNormal
from ..lazy import Flow, LazyTransform, UnconditionalDistribution
from ..nn import MLP
from ..transforms import CouplingTransform, DependentTransform, MonotonicAffineTransform
from ..utils import broadcast, unpack


class GeneralCouplingTransform(LazyTransform):
    r"""Creates a lazy general coupling transformation.

    See also:
        :class:`zuko.transforms.CouplingTransform`

    References:
        | NICE: Non-linear Independent Components Estimation (Dinh et al., 2014)
        | https://arxiv.org/abs/1410.8516

    Arguments:
        features: The number of features.
        context: The number of context features.
        mask: The coupling mask. If :py:`None`, use a checkered mask.
        univariate: The univariate transformation constructor.
        shapes: The shapes of the univariate transformation parameters.
        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.

    Example:
        >>> t = GeneralCouplingTransform(3, 4)
        >>> t
        GeneralCouplingTransform(
          (base): MonotonicAffineTransform()
          (mask): [0, 1, 0]
          (hyper): MLP(
            (0): Linear(in_features=5, out_features=64, bias=True)
            (1): ReLU()
            (2): Linear(in_features=64, out_features=64, bias=True)
            (3): ReLU()
            (4): Linear(in_features=64, out_features=4, bias=True)
          )
        )
        >>> x = torch.randn(3)
        >>> x
        tensor([-0.7900, -0.3259, -1.3184])
        >>> c = torch.randn(4)
        >>> y = t(c)(x)
        >>> t(c).inv(y)
        tensor([-0.7900, -0.3259, -1.3184], grad_fn=<IndexPutBackward0>)
    """

    def __new__(
        cls,
        features: int = None,
        context: int = 0,
        mask: BoolTensor = None,
        *args,
        **kwargs,
    ) -> LazyTransform:
        if features is None or features > 1:
            return super().__new__(cls)
        else:
            return ElementWiseTransform(features, context, *args, **kwargs)

    def __init__(
        self,
        features: int,
        context: int = 0,
        mask: BoolTensor = None,
        univariate: Callable[..., Transform] = MonotonicAffineTransform,
        shapes: Sequence[Size] = ((), ()),
        **kwargs,
    ):
        super().__init__()

        # Univariate transformation
        self.univariate = univariate
        self.shapes = shapes
        self.total = sum(prod(s) for s in shapes)

        # Mask
        if mask is None:
            mask = torch.arange(features) % 2 == 1
        else:
            mask = torch.as_tensor(mask, dtype=bool)

        assert mask.ndim == 1, "'mask' should be a vector."
        assert mask.shape[0] == features, f"'mask' should have {features} elements."

        features_a = mask.sum().item()
        features_b = features - features_a

        assert features_a > 0
        assert features_b > 0

        self.register_buffer("mask", mask)

        # Hyper network
        self.hyper = MLP(features_a + context, features_b * self.total, **kwargs)

    def extra_repr(self) -> str:
        base = self.univariate(*map(torch.randn, self.shapes))
        mask = self.mask.int().tolist()

        if len(mask) > 10:
            mask = mask[:5] + [...] + mask[-5:]
            mask = str(mask).replace("Ellipsis", "...")

        return "\n".join([
            f"(base): {base}",
            f"(mask): {mask}",
        ])

    def meta(self, c: Tensor, x: Tensor) -> Transform:
        if c is not None:
            x = torch.cat(broadcast(x, c, ignore=1), dim=-1)

        phi = self.hyper(x)
        phi = phi.unflatten(-1, (-1, self.total))
        phi = unpack(phi, self.shapes)

        return DependentTransform(self.univariate(*phi), 1)

    def forward(self, c: Tensor = None) -> Transform:
        return CouplingTransform(partial(self.meta, c), self.mask)


class NICE(Flow):
    r"""Creates a NICE / RealNVP flow.

    Affine transformations are used by default, as in RealNVP (Dinh et al., 2016),
    instead of additive transformations, as in NICE (Dinh et al., 2014).

    References:
        | NICE: Non-linear Independent Components Estimation (Dinh et al., 2014)
        | https://arxiv.org/abs/1410.8516

        | Density estimation using Real NVP (Dinh et al., 2016)
        | https://arxiv.org/abs/1605.08803

    Arguments:
        features: The number of features.
        context: The number of context features.
        transforms: The number of coupling transformations.
        randmask: Whether random coupling masks are used or not. If :py:`False`,
            use alternating checkered masks.
        kwargs: Keyword arguments passed to :class:`GeneralCouplingTransform`.
    """

    def __init__(
        self,
        features: int,
        context: int = 0,
        transforms: int = 3,
        randmask: bool = False,
        **kwargs,
    ):
        temp = []

        for i in range(transforms):
            if randmask:
                mask = torch.randperm(features) % 2 == i % 2
            else:
                mask = torch.arange(features) % 2 == i % 2

            temp.append(
                GeneralCouplingTransform(
                    features=features,
                    context=context,
                    mask=mask,
                    **kwargs,
                )
            )

        base = UnconditionalDistribution(
            DiagNormal,
            torch.zeros(features),
            torch.ones(features),
            buffer=True,
        )

        super().__init__(temp, base)


class RealNVP(NICE):
    pass



r"""Doctests configuration."""

import pytest
import torch

import zuko


@pytest.fixture(autouse=True, scope="module")
def doctest_imports(doctest_namespace):
    doctest_namespace["torch"] = torch
    doctest_namespace["zuko"] = zuko


@pytest.fixture(autouse=True)
def torch_seed():
    with torch.random.fork_rng():
        yield torch.random.manual_seed(0)



r"""Core building blocks.

Warning:
    This sub-module is deprecated and will be removed in the future. Use
    :mod:`zuko.lazy` instead.
"""

__all__ = [
    "Flow",
]

from ..lazy import (  # noqa: F401
    Flow,
    LazyComposedTransform,
    LazyDistribution,
    LazyInverse,
    LazyTransform,
    Unconditional,
    UnconditionalDistribution,
    UnconditionalTransform,
)



r"""Parameterizable transformations."""

__all__ = [
    "AdditiveTransform",
    "AutoregressiveTransform",
    "BernsteinTransform",
    "BoundedBernsteinTransform",
    "CircularShiftTransform",
    "ComposedTransform",
    "CosTransform",
    "CouplingTransform",
    "DependentTransform",
    "FreeFormJacobianTransform",
    "GaussianizationTransform",
    "IdentityTransform",
    "LULinearTransform",
    "MonotonicAffineTransform",
    "MonotonicRQSTransform",
    "MonotonicTransform",
    "PermutationTransform",
    "RotationTransform",
    "SOSPolynomialTransform",
    "SignedPowerTransform",
    "SinTransform",
    "SoftclipTransform",
    "UnconstrainedMonotonicTransform",
]

import math
import torch
import torch.nn.functional as F

from textwrap import indent
from torch import BoolTensor, LongTensor, Size, Tensor
from torch.distributions import Distribution, Transform, constraints
from torch.distributions.transforms import *  # noqa: F403
from torch.distributions.utils import _sum_rightmost
from typing import Any, Callable, Iterable, Tuple, Union

from .utils import bisection, broadcast, gauss_legendre, odeint

torch.distributions.transforms._InverseTransform.__name__ = "Inverse"


def _call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
    r"""Returns both the transformed value and the log absolute determinant of the
    transformation's Jacobian."""

    y = self.__call__(x)
    ladj = self.log_abs_det_jacobian(x, y)

    return y, ladj


Transform.call_and_ladj = _call_and_ladj


class ComposedTransform(Transform):
    r"""Creates a transformation :math:`f(x) = f_n \circ \dots \circ f_0(x)`.

    Optimized version of :class:`torch.distributions.transforms.ComposeTransform`.

    Arguments:
        transforms: A sequence of transformations :math:`f_i`.
    """

    def __init__(self, *transforms: Transform, **kwargs):
        super().__init__(**kwargs)

        assert transforms, "'transforms' cannot be empty"

        event_dim = 0

        for t in reversed(transforms):
            event_dim = t.domain.event_dim + max(event_dim - t.codomain.event_dim, 0)

        self.domain_dim = event_dim

        for t in transforms:
            event_dim += t.codomain.event_dim - t.domain.event_dim

        self.codomain_dim = event_dim
        self.transforms = transforms

    def __repr__(self) -> str:
        lines = [f"({i}): {t}" for i, t in enumerate(self.transforms)]
        lines = indent("\n".join(lines), "  ")

        return f"{self.__class__.__name__}(\n" + lines + "\n)"

    @property
    def domain(self) -> constraints.Constraint:
        domain = self.transforms[0].domain
        reinterpreted = self.domain_dim - domain.event_dim

        if reinterpreted > 0:
            return constraints.independent(domain, reinterpreted)
        else:
            return domain

    @property
    def codomain(self) -> constraints.Constraint:
        codomain = self.transforms[-1].codomain
        reinterpreted = self.codomain_dim - codomain.event_dim

        if reinterpreted > 0:
            return constraints.independent(codomain, reinterpreted)
        else:
            return codomain

    @property
    def bijective(self) -> bool:
        return all(t.bijective for t in self.transforms)

    def _call(self, x: Tensor) -> Tensor:
        for t in self.transforms:
            x = t(x)
        return x

    @property
    def inv(self) -> Transform:
        new = self.__new__(ComposedTransform)
        new.transforms = [t.inv for t in reversed(self.transforms)]
        new.domain_dim = self.codomain_dim
        new.codomain_dim = self.domain_dim

        Transform.__init__(new)

        return new

    def _inverse(self, y: Tensor) -> Tensor:
        for t in reversed(self.transforms):
            y = t.inv(y)
        return y

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        _, ladj = self.call_and_ladj(x)
        return ladj

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        event_dim = self.domain_dim
        acc = 0

        for t in self.transforms:
            x, ladj = t.call_and_ladj(x)
            acc = acc + _sum_rightmost(ladj, event_dim - t.domain.event_dim)
            event_dim += t.codomain.event_dim - t.domain.event_dim

        return x, acc

    def forward_shape(self, shape: Size) -> Size:
        for t in self.transforms:
            shape = t.forward_shape(shape)
        return shape

    def inverse_shape(self, shape: Size) -> Size:
        for t in reversed(self.transforms):
            shape = t.inverse_shape(shape)
        return shape


class DependentTransform(Transform):
    r"""Wraps a base transformation to treat right-most dimensions as dependent.

    Optimized version of :class:`torch.distributions.transforms.IndependentTransform`.

    Arguments:
        base: The base transformation.
        reinterpreted: The number of dimensions to treat as dependent.
    """

    def __init__(self, base: Transform, reinterpreted: int, **kwargs):
        super().__init__(**kwargs)

        self.base = base
        self.reinterpreted = reinterpreted

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base}, {self.reinterpreted})"

    @property
    def domain(self) -> constraints.Constraint:
        return constraints.independent(self.base.domain, self.reinterpreted)

    @property
    def codomain(self) -> constraints.Constraint:
        return constraints.independent(self.base.codomain, self.reinterpreted)

    @property
    def bijective(self) -> bool:
        return self.base.bijective

    def _call(self, x: Tensor) -> Tensor:
        return self.base(x)

    @property
    def inv(self) -> Transform:
        return DependentTransform(self.base.inv, self.reinterpreted)

    def _inverse(self, y: Tensor) -> Tensor:
        return self.base.inv(y)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        ladj = self.base.log_abs_det_jacobian(x, y)
        ladj = _sum_rightmost(ladj, self.reinterpreted)

        return ladj

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        y, ladj = self.base.call_and_ladj(x)
        ladj = ladj = _sum_rightmost(ladj, self.reinterpreted)

        return y, ladj

    def forward_shape(self, shape: Size) -> Size:
        return self.base.forward_shape(shape)

    def inverse_shape(self, shape: Size) -> Size:
        return self.base.inverse_shape(shape)


class IdentityTransform(Transform):
    r"""Creates a transformation :math:`f(x) = x`."""

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, IdentityTransform)

    def _call(self, x: Tensor) -> Tensor:
        return x

    def _inverse(self, y: Tensor) -> Tensor:
        return y

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.zeros_like(x)


class CosTransform(Transform):
    r"""Creates a transformation :math:`f(x) = -\cos(x)`."""

    domain = constraints.interval(0, math.pi)
    codomain = constraints.interval(-1, 1)
    bijective = True
    sign = +1

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, CosTransform)

    def _call(self, x: Tensor) -> Tensor:
        return -x.cos()

    def _inverse(self, y: Tensor) -> Tensor:
        return (-y).acos()

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return x.sin().abs().log()


class SinTransform(Transform):
    r"""Creates a transformation :math:`f(x) = \sin(x)`."""

    domain = constraints.interval(-math.pi / 2, math.pi / 2)
    codomain = constraints.interval(-1, 1)
    bijective = True
    sign = +1

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SinTransform)

    def _call(self, x: Tensor) -> Tensor:
        return x.sin()

    def _inverse(self, y: Tensor) -> Tensor:
        return y.asin()

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return x.cos().abs().log()


class SoftclipTransform(Transform):
    r"""Creates a transformation that maps :math:`\mathbb{R}` to the interval
    :math:`[-B, B]`.

    .. math:: f(x) = \frac{x}{1 + \left| \frac{x}{B} \right|}

    Arguments:
        bound: The codomain bound :math:`B`.
    """

    bijective = True
    sign = +1

    def __init__(self, bound: float = 1.0, **kwargs):
        super().__init__(**kwargs)

        self.bound = bound
        self.domain = constraints.real
        self.codomain = constraints.interval(-bound, bound)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bound={self.bound})"

    def _call(self, x: Tensor) -> Tensor:
        return x / (1 + abs(x / self.bound))

    def _inverse(self, y: Tensor) -> Tensor:
        return y / (1 - abs(y / self.bound))

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return -2 * torch.log1p(abs(x / self.bound))


class CircularShiftTransform(Transform):
    r"""Creates a transformation that circularly shifts the interval :math:`[-B, B]`.

    .. math:: f(x) = (x \bmod 2B) - B

    Note:
        This transformation is only bijective over its domain :math:`[-B, B]` as
        :math:`f(x) = f(x + 2kB)` for all :math:`k \in \mathbb{Z}`.

    Arguments:
        bound: The domain bound :math:`B`.
    """

    bijective = True

    def __init__(self, bound: float = 1.0, **kwargs):
        super().__init__(**kwargs)

        self.bound = bound
        self.domain = constraints.interval(-bound, bound)
        self.codomain = constraints.interval(-bound, bound)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bound={self.bound})"

    def _call(self, x: Tensor) -> Tensor:
        return torch.remainder(x, 2 * self.bound) - self.bound

    def _inverse(self, y: Tensor) -> Tensor:
        return torch.remainder(y, 2 * self.bound) - self.bound

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.zeros_like(x)


class SignedPowerTransform(Transform):
    r"""Creates a transformation :math:`f(x) = \mathrm{sign}(x) |x|^{\exp(\alpha)}`.

    Arguments:
        alpha: The unconstrained exponent :math:`\alpha`, with shape :math:`(*,)`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(self, alpha: Tensor, **kwargs):
        super().__init__(**kwargs)

        self.alpha = alpha

    def _call(self, x: Tensor) -> Tensor:
        return x * abs(x) ** torch.expm1(self.alpha)

    def _inverse(self, y: Tensor) -> Tensor:
        return y * abs(y) ** torch.expm1(-self.alpha)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return self.alpha + torch.expm1(self.alpha) * torch.log(abs(x))


class AdditiveTransform(Transform):
    r"""Creates a transformation :math:`f(x) = x + b`.

    References:
        | NICE: Non-linear Independent Components Estimation (Dinh et al., 2014)
        | https://arxiv.org/abs/1410.8516

    Arguments:
        shift: The shift term :math:`b`, with shape :math:`(*,)`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(self, shift: Tensor, **kwargs):
        super().__init__(**kwargs)

        self.shift = shift

    def _call(self, x: Tensor) -> Tensor:
        return x + self.shift

    def _inverse(self, y: Tensor) -> Tensor:
        return y - self.shift

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.zeros_like(x)


class MonotonicAffineTransform(Transform):
    r"""Creates a transformation :math:`f(x) = \exp(a) x + b`.

    Arguments:
        shift: The shift term :math:`b`, with shape :math:`(*,)`.
        scale: The unconstrained scale factor :math:`a`, with shape :math:`(*,)`.
        slope: The minimum slope of the transformation.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(
        self,
        shift: Tensor,
        scale: Tensor,
        slope: float = 1e-3,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.shift = shift
        self.log_scale = scale / (1 + abs(scale / math.log(slope)))
        self.scale = self.log_scale.exp()

    def _call(self, x: Tensor) -> Tensor:
        return x * self.scale + self.shift

    def _inverse(self, y: Tensor) -> Tensor:
        return (y - self.shift) / self.scale

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return self.log_scale.expand(x.shape)


class MonotonicRQSTransform(Transform):
    r"""Creates a monotonic rational-quadratic spline (RQS) transformation.

    References:
        | Neural Spline Flows (Durkan et al., 2019)
        | https://arxiv.org/abs/1906.04032

    Arguments:
        widths: The unconstrained bin widths, with shape :math:`(*, K)`.
        heights: The unconstrained bin heights, with shape :math:`(*, K)`.
        derivatives: The unconstrained knot derivatives, with shape :math:`(*, K - 1)`.
        bound: The spline's (co)domain bound :math:`B`.
        slope: The minimum slope of the transformation.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(
        self,
        widths: Tensor,
        heights: Tensor,
        derivatives: Tensor,
        bound: float = 5.0,
        slope: float = 1e-3,
        **kwargs,
    ):
        super().__init__(**kwargs)

        widths = widths / (1 + abs(2 * widths / math.log(slope)))
        heights = heights / (1 + abs(2 * heights / math.log(slope)))
        derivatives = derivatives / (1 + abs(derivatives / math.log(slope)))

        widths = F.pad(F.softmax(widths, dim=-1), (1, 0), value=0)
        heights = F.pad(F.softmax(heights, dim=-1), (1, 0), value=0)
        derivatives = F.pad(derivatives, (1, 1), value=0)

        self.horizontal = bound * (2 * torch.cumsum(widths, dim=-1) - 1)
        self.vertical = bound * (2 * torch.cumsum(heights, dim=-1) - 1)
        self.derivatives = torch.exp(derivatives)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(bins={self.bins})"

    @property
    def bins(self) -> int:
        return self.horizontal.shape[-1] - 1

    def bin(self, k: LongTensor) -> Tuple[Tensor, ...]:
        mask = torch.logical_and(0 <= k, k < self.bins)

        k = k % self.bins
        k0_k1 = torch.stack((k, k + 1))

        k0_k1, hs, vs, ds = broadcast(
            k0_k1[..., None],
            self.horizontal,
            self.vertical,
            self.derivatives,
            ignore=1,
        )

        x0, x1 = hs.gather(-1, k0_k1).squeeze(dim=-1)
        y0, y1 = vs.gather(-1, k0_k1).squeeze(dim=-1)
        d0, d1 = ds.gather(-1, k0_k1).squeeze(dim=-1)

        s = (y1 - y0) / (x1 - x0)

        return mask, x0, x1, y0, y1, d0, d1, s

    @staticmethod
    def searchsorted(seq: Tensor, value: Tensor) -> LongTensor:
        return torch.sum(seq < value[..., None], dim=-1)

    def _call(self, x: Tensor) -> Tensor:
        k = self.searchsorted(self.horizontal, x) - 1
        mask, x0, x1, y0, y1, d0, d1, s = self.bin(k)

        z = mask * (x - x0) / (x1 - x0)
        y = y0 + (y1 - y0) * (s * z**2 + d0 * z * (1 - z)) / (s + (d0 + d1 - 2 * s) * z * (1 - z))

        return torch.where(mask, y, x)

    def _inverse(self, y: Tensor) -> Tensor:
        k = self.searchsorted(self.vertical, y) - 1
        mask, x0, x1, y0, y1, d0, d1, s = self.bin(k)

        y_ = mask * (y - y0)

        a = (y1 - y0) * (s - d0) + y_ * (d0 + d1 - 2 * s)
        b = (y1 - y0) * d0 - y_ * (d0 + d1 - 2 * s)
        c = -s * y_

        z = 2 * c / (-b - (b**2 - 4 * a * c).sqrt())

        x = x0 + z * (x1 - x0)

        return torch.where(mask, x, y)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        _, ladj = self.call_and_ladj(x)
        return ladj

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        k = self.searchsorted(self.horizontal, x) - 1
        mask, x0, x1, y0, y1, d0, d1, s = self.bin(k)

        z = mask * (x - x0) / (x1 - x0)
        y = y0 + (y1 - y0) * (s * z**2 + d0 * z * (1 - z)) / (s + (d0 + d1 - 2 * s) * z * (1 - z))

        jacobian = (
            s**2
            * (2 * s * z * (1 - z) + d0 * (1 - z) ** 2 + d1 * z**2)
            / (s + (d0 + d1 - 2 * s) * z * (1 - z)) ** 2
        )

        return torch.where(mask, y, x), mask * jacobian.log()


class MonotonicTransform(Transform):
    r"""Creates a transformation from a monotonic univariate function :math:`f_\phi(x)`.

    The inverse function :math:`f_\phi^{-1}` is approximated using the bisection method.

    Arguments:
        f: A monotonic univariate function :math:`f_\phi`. If :py:`None`, :py:`self.f`
            is used instead.
        phi: The parameters :math:`\phi` of :math:`f_\phi`. Providing the parameters
            is required to make the inverse transformation trainable.
        bound: The domain bound :math:`B`.
        eps: The absolute tolerance for the inverse transformation.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(
        self,
        f: Callable[[Tensor], Tensor] = None,
        phi: Iterable[Tensor] = (),
        bound: float = 10.0,
        eps: float = 1e-6,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if f is not None:
            self.f = f

        self.phi = tuple(filter(lambda p: p.requires_grad, phi))
        self.bound = bound
        self.eps = eps

    def _call(self, x: Tensor) -> Tensor:
        return self.f(x)

    def _inverse(self, y: Tensor) -> Tensor:
        return bisection(
            f=self.f,
            y=y,
            a=torch.full_like(y, -self.bound),
            b=torch.full_like(y, self.bound),
            n=math.ceil(math.log2(2 * self.bound / self.eps)),
            phi=self.phi,
        )

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        _, ladj = self.call_and_ladj(x)
        return ladj

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        with torch.enable_grad():
            x = x.clone().requires_grad_()
            y = self.f(x)

        jacobian = torch.autograd.grad(y, x, torch.ones_like(y), create_graph=True)[0]

        return y, jacobian.log()


class BernsteinTransform(MonotonicTransform):
    r"""Creates a monotonic Bernstein polynomial transformation.

    .. math:: f(x) = \frac{1}{M + 1} \sum_{i=0}^{M} b_{i+1,M-i+1} \left( \frac{x + B}{2B} \right) \, \theta_i

    where :math:`b_{i,j}` are the Bernstein basis polynomials.

    As the polynomial :math:`f(x)` is only defined for :math:`x \in [-B, B]`, the
    transformation linearly extrapolates it outside this domain. The second derivative
    at the bounds is enforced to be zero for smooth extrapolation.

    References:
        | Deep transformation models: Tackling complex regression problems with neural network based transformation models (Sick et al., 2020)
        | https://arxiv.org/abs/2004.00464

        | Short-Term Density Forecasting of Low-Voltage Load using Bernstein-Polynomial Normalizing Flows (Arpogaus et al., 2022)
        | https://arxiv.org/abs/2204.13939

    Wikipedia:
        https://wikipedia.org/wiki/Bernstein_polynomial

    Arguments:
        theta: The unconstrained polynomial coefficients :math:`\theta`,
            with shape :math:`(*, M - 2)`.
        bound: The polynomial's domain bound :math:`B`.
        kwargs: Keyword arguments passed to :class:`MonotonicTransform`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(self, theta: Tensor, bound: float = 5.0, **kwargs):
        super().__init__(None, phi=(theta,), bound=bound, **kwargs)

        self.theta = self._constrain_theta(theta)
        self.basis = self._bernstein_basis(self.order, device=theta.device, dtype=theta.dtype)

        self.offset, self.slope = self._offset_and_slope()

    @property
    def order(self) -> int:
        return self.theta.shape[-1] - 1

    def _offset_and_slope(self) -> Tuple[Tuple[Tensor, Tensor], Tuple[Tensor, Tensor]]:
        r"""Calculates the offsets and slopes at the domain bounds for extrapolation."""

        dtheta = self.order * (self.theta[..., 1:] - self.theta[..., :-1])
        dbasis = self._bernstein_basis(
            self.order - 1, device=self.theta.device, dtype=self.theta.dtype
        )

        bounds = [
            self.theta.new_tensor(self.eps),
            self.theta.new_tensor(1 - self.eps),
        ]

        offset = [self._bernstein_poly(x, self.theta, self.basis) for x in bounds]
        slope = [self._bernstein_poly(x, dtheta, dbasis) for x in bounds]

        return tuple(offset), tuple(slope)

    @staticmethod
    def _constrain_theta(unconstrained_theta: Tensor) -> Tensor:
        """Processes the unconstrained output of the hyper-network to be increasing."""

        shift = math.log(2.0) * unconstrained_theta.shape[-1] / 2

        theta_min = unconstrained_theta[..., :1]
        unconstrained_theta = unconstrained_theta[..., 1:]

        # ensure smooth bounds
        unconstrained_theta = torch.cat(
            (
                unconstrained_theta[..., :1],
                unconstrained_theta,
                unconstrained_theta[..., -1:],
            ),
            dim=-1,
        )

        diffs = torch.nn.functional.softplus(unconstrained_theta)
        diffs = torch.cat((theta_min, diffs), dim=-1)

        theta = torch.cumsum(diffs, dim=-1) - shift

        return theta

    @staticmethod
    def _bernstein_basis(order: int, **kwargs) -> Distribution:
        alpha = torch.arange(1, order + 2, **kwargs)
        beta = torch.arange(order + 1, 0, -1, **kwargs)
        basis = torch.distributions.Beta(alpha, beta)
        return basis

    @staticmethod
    def _bernstein_poly(x: Tensor, theta: Tensor, basis: Distribution) -> Tensor:
        b = basis.log_prob(x.unsqueeze(-1)).exp()
        y = torch.mean(b * theta, dim=-1)
        return y

    def f(self, x: Tensor) -> Tensor:
        x = (x + self.bound) / (2 * self.bound)  # map [-B, B] to [0, 1]

        lower_bound = x <= self.eps
        upper_bound = x >= 1 - self.eps
        x_safe = torch.where(lower_bound | upper_bound, 0.5 * torch.ones_like(x), x)

        y = self._bernstein_poly(x_safe, self.theta, self.basis)

        # f'(eps) * (x - eps) + f(eps)
        y0 = self.slope[0] * (x - self.eps) + self.offset[0]

        # f'(1-eps) * (x - 1 - eps) + f(1-eps)
        y1 = self.slope[1] * (x - 1 + self.eps) + self.offset[1]

        y = torch.where(lower_bound, y0, y)
        y = torch.where(upper_bound, y1, y)

        return y

    def _inverse(self, y: Tensor) -> Tensor:
        left_bound = y <= self.offset[0]
        right_bound = y >= self.offset[1]

        x = super()._inverse(y)
        x0 = (y - self.offset[0]) / self.slope[0] + self.eps
        x1 = (y - self.offset[1]) / self.slope[1] - self.eps + 1

        # map [0, 1] to [-B, B]
        x0 = x0 * 2 * self.bound - self.bound
        x1 = x1 * 2 * self.bound - self.bound

        x = torch.where(left_bound, x0, x)
        x = torch.where(right_bound, x1, x)

        return x


class BoundedBernsteinTransform(BernsteinTransform):
    r"""Creates a bounded Bernstein polynomial transformation.

    This subclass of :class:`BernsteinTransform` scales the Bernstein coefficients so
    that the polynomial's domain and codomain match the interval :math:`[-B, B]`. It
    also enforces that the first derivative at the bounds is one and that the second
    derivative is zero, ensuring a smooth transition to the identity function outside
    the bounds.

    These constraints make the transformation suitable for chaining in flows.

    Arguments:
        theta: The unconstrained polynomial coefficients :math:`\theta`,
            with shape :math:`(*, M - 5)`.
        kwargs: Keyword arguments passed to :class:`BernsteinTransform`.
    """

    def _constrain_theta(self, unconstrained_theta: Tensor) -> Tensor:
        theta_min = -self.bound * torch.ones_like(unconstrained_theta[..., :1])

        diff_on_bounds = (2 * self.bound) / (unconstrained_theta.shape[-1] + 4)
        diffs = torch.nn.functional.softmax(unconstrained_theta, dim=-1) * (
            2 * self.bound - 4 * diff_on_bounds
        )

        # ensure identity on bounds by enforcing Be'(0,1) = 1 and Be''(0,1) = 0
        # Be'(0) = order * theta_1 - theta_0 = order * diff_0 -> diff_0 = 1 / order
        # Be''(0) = (order - 1) * (diff_1 - diff_0) -> diff_0 = diff_1
        diffs = torch.cat(
            (
                theta_min,
                diff_on_bounds * torch.ones_like(diffs[..., :2]),
                diffs,
                diff_on_bounds * torch.ones_like(diffs[..., :2]),
            ),
            dim=-1,
        )

        return torch.cumsum(diffs, dim=-1)

    def _offset_and_slope(self) -> Tuple[Tuple[Tensor, Tensor], Tuple[Tensor, Tensor]]:
        offset = (
            self.theta.new_tensor(-self.bound),
            self.theta.new_tensor(self.bound),
        )

        slope = (
            self.theta.new_tensor(2 * self.bound),
            self.theta.new_tensor(2 * self.bound),
        )

        return offset, slope


class GaussianizationTransform(MonotonicTransform):
    r"""Creates a gaussianization transformation.

    .. math:: f(x) = \Phi^{-1}
        \left( \frac{1}{K} \sum_{i=1}^K \Phi(\exp(a_i) x + b_i) \right)

    where :math:`\Phi` is the cumulative distribution function (CDF) of the standard
    normal :math:`\mathcal{N}(0, 1)`.

    References:
        | Gaussianization (Chen et al., 2000)
        | https://papers.nips.cc/paper/1856-gaussianization

    Arguments:
        shift: The shift terms :math:`b`, with shape :math:`(*, K)`.
        scale: The unconstrained scale factors :math:`a`, with shape :math:`(*, K)`.
        kwargs: Keyword arguments passed to :class:`MonotonicTransform`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(
        self,
        shift: Tensor,
        scale: Tensor,
        **kwargs,
    ):
        super().__init__(None, phi=(shift, scale), **kwargs)

        self.shift = shift
        self.scale = torch.exp(scale)

    def f(self, x: Tensor) -> Tensor:
        x = x[..., None] * self.scale + self.shift
        x = torch.erf(x / math.sqrt(2))
        x = torch.mean(x, dim=-1) * (1 - 1e-6)
        x = torch.erfinv(x) * math.sqrt(2)

        return x


class UnconstrainedMonotonicTransform(MonotonicTransform):
    r"""Creates a monotonic transformation :math:`f(x)` by integrating a positive
    univariate function :math:`g(x)`.

    .. math:: f(x) = \int_0^x g(u) ~ du

    The definite integral is estimated by a :math:`n`-point Gauss-Legendre quadrature.

    Arguments:
        g: A positive univariate function :math:`g`. If :py:`None`, :py:`self.g` is
            used instead.
        n: The number of points :math:`n` for the quadrature.
        kwargs: Keyword arguments passed to :class:`MonotonicTransform`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(
        self,
        g: Callable[[Tensor], Tensor] = None,
        n: int = 32,
        **kwargs,
    ):
        super().__init__(None, **kwargs)

        if g is not None:
            self.g = g

        self.n = n

    def f(self, x: Tensor) -> Tensor:
        return gauss_legendre(
            f=self.g,
            a=torch.zeros_like(x),
            b=x,
            n=self.n,
            phi=self.phi,
        )

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return self.g(x).log()

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        return self.f(x), self.g(x).log()


class SOSPolynomialTransform(UnconstrainedMonotonicTransform):
    r"""Creates a sum-of-squares (SOS) polynomial transformation.

    The transformation :math:`f(x)` is expressed as the integral of the
    sum of :math:`K` squared polynomials of degree :math:`L`.

    .. math:: f(x) = \int_0^x \frac{1}{K} \sum_{i = 1}^K
        \left( 1 + \sum_{j = 0}^L a_{i,j} ~ u^j \right)^2 ~ du

    References:
        | Sum-of-Squares Polynomial Flow (Jaini et al., 2019)
        | https://arxiv.org/abs/1905.02325

    Arguments:
        a: The polynomial coefficients :math:`a`, with shape :math:`(*, K, L + 1)`.
        slope: The minimum slope of the transformation.
        kwargs: Keyword arguments passed to :class:`UnconstrainedMonotonicTransform`.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True
    sign = +1

    def __init__(self, a: Tensor, slope: float = 1e-3, **kwargs):
        super().__init__(None, phi=(a,), n=a.shape[-1], **kwargs)

        self.a = a
        self.i = torch.arange(a.shape[-1], device=a.device)
        self.slope = slope

    def g(self, x: Tensor) -> Tensor:
        x = x / self.bound
        x = x[..., None] ** self.i
        p = 1 + self.a @ x[..., None]

        return p.squeeze(dim=-1).square().mean(dim=-1) + self.slope


class AutoregressiveTransform(Transform):
    r"""Transform via an autoregressive scheme.

    .. math:: y_i = f(x_i | x_{<i})

    Arguments:
        meta: A function which returns a transformation :math:`f` given :math:`x`.
        passes: The number of passes for the inverse transformation.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(
        self,
        meta: Callable[[Tensor], Transform],
        passes: int,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.meta = meta
        self.passes = passes

    def _call(self, x: Tensor) -> Tensor:
        return self.meta(x)(x)

    def _inverse(self, y: Tensor) -> Tensor:
        x = torch.zeros_like(y)
        for _ in range(self.passes):
            x = self.meta(x).inv(y)

        return x

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return self.meta(x).log_abs_det_jacobian(x, y)

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        y, ladj = self.meta(x).call_and_ladj(x)
        return y, ladj


class CouplingTransform(Transform):
    r"""Transform via a coupling scheme.

    .. math::
        y_a & = x_a \\
        y_b & = f(x_b | x_a)

    Arguments:
        meta: A function which returns a transformation :math:`f` given :math:`x_a`.
        mask: A coupling mask defining the split :math:`x \to (x_a, x_b)`. Ones
            correspond to the constant split :math:`x_a` and zeros to the transformed
            split :math:`x_b`.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(
        self,
        meta: Callable[[Tensor], Transform],
        mask: BoolTensor,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.meta = meta
        self.idx_a = mask.nonzero().squeeze(-1)
        self.idx_b = (~mask).nonzero().squeeze(-1)

    def split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        return x[..., self.idx_a], x[..., self.idx_b]

    def merge(self, x_a: Tensor, x_b: Tensor, shape: Size) -> Tensor:
        x = x_a.new_empty(shape)
        x[..., self.idx_a] = x_a
        x[..., self.idx_b] = x_b

        return x

    def _call(self, x: Tensor) -> Tensor:
        x_a, x_b = self.split(x)
        y_b = self.meta(x_a)(x_b)

        return self.merge(x_a, y_b, x.shape)

    def _inverse(self, y: Tensor) -> Tensor:
        y_a, y_b = self.split(y)
        x_b = self.meta(y_a).inv(y_b)

        return self.merge(y_a, x_b, y.shape)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        x_a, x_b = self.split(x)
        _, y_b = self.split(y)

        return self.meta(x_a).log_abs_det_jacobian(x_b, y_b)

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        x_a, x_b = self.split(x)
        y_b, ladj = self.meta(x_a).call_and_ladj(x_b)
        y = self.merge(x_a, y_b, x.shape)

        return y, ladj


class FreeFormJacobianTransform(Transform):
    r"""Creates a free-form Jacobian transformation.

    The transformation is the integration of a system of first-order ordinary
    differential equations

    .. math:: x(t_1) = x_0 + \int_{t_0}^{t_1} f_\phi(t, x(t)) ~ dt .

    References:
        | FFJORD: Free-form Continuous Dynamics for Scalable Reversible Generative Models (Grathwohl et al., 2018)
        | https://arxiv.org/abs/1810.01367

    Arguments:
        f: A system of first-order ODEs :math:`f_\phi`.
        t0: The initial integration time :math:`t_0`.
        t1: The final integration time :math:`t_1`.
        phi: The parameters :math:`\phi` of :math:`f_\phi`.
        atol: The absolute integration tolerance.
        rtol: The relative integration tolerance.
        exact: Whether the exact log-determinant of the Jacobian or an unbiased
            stochastic estimate thereof is calculated.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(
        self,
        f: Callable[[Tensor, Tensor], Tensor],
        t0: Union[float, Tensor] = 0.0,
        t1: Union[float, Tensor] = 1.0,
        phi: Iterable[Tensor] = (),
        atol: float = 1e-6,
        rtol: float = 1e-5,
        exact: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.f = f
        self.t0 = t0
        self.t1 = t1
        self.phi = tuple(filter(lambda p: p.requires_grad, phi))
        self.atol = atol
        self.rtol = rtol
        self.exact = exact
        self.trace_scale = 1e-2  # relax jacobian tolerances

    def _call(self, x: Tensor) -> Tensor:
        return odeint(self.f, x, self.t0, self.t1, self.phi, self.atol, self.rtol)

    @property
    def inv(self) -> Transform:
        return FreeFormJacobianTransform(
            f=self.f,
            t0=self.t1,
            t1=self.t0,
            phi=self.phi,
            atol=self.atol,
            rtol=self.rtol,
            exact=self.exact,
        )

    def _inverse(self, y: Tensor) -> Tensor:
        return odeint(self.f, y, self.t1, self.t0, self.phi, self.atol, self.rtol)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        _, ladj = self.call_and_ladj(x)
        return ladj

    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        if self.exact:
            I = torch.eye(x.shape[-1], dtype=x.dtype, device=x.device)
            I = I.expand(*x.shape, -1).movedim(-1, 0)
        else:
            eps = torch.randn_like(x)

        def f_aug(t: Tensor, x: Tensor, ladj: Tensor) -> Tensor:
            with torch.enable_grad():
                x = x.clone().requires_grad_()
                dx = self.f(t, x)

            if self.exact:
                (jacobian,) = torch.autograd.grad(
                    dx, x, I, create_graph=True, is_grads_batched=True
                )
                trace = torch.einsum("i...i", jacobian)
            else:
                (epsjp,) = torch.autograd.grad(dx, x, eps, create_graph=True)
                trace = (epsjp * eps).sum(dim=-1)

            return dx, trace * self.trace_scale

        ladj = torch.zeros_like(x[..., 0])
        y, ladj = odeint(f_aug, (x, ladj), self.t0, self.t1, self.phi, self.atol, self.rtol)

        return y, ladj * (1 / self.trace_scale)


class PermutationTransform(Transform):
    r"""Creates a transformation that permutes the elements.

    Arguments:
        order: The permutation order, with shape :math:`(*, D)`.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, order: LongTensor, **kwargs):
        super().__init__(**kwargs)

        self.order = order

    def __repr__(self) -> str:
        order = self.order.tolist()

        if len(order) > 10:
            order = order[:5] + [...] + order[-5:]
            order = str(order).replace("Ellipsis", "...")

        return f"{self.__class__.__name__}({order})"

    def _call(self, x: Tensor) -> Tensor:
        return x[..., self.order]

    def _inverse(self, y: Tensor) -> Tensor:
        return y[..., torch.argsort(self.order)]

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.zeros_like(x[..., 0])


class RotationTransform(Transform):
    r"""Creates a rotation transformation :math:`f(x) = R x`.

    .. math:: R = \exp(A - A^T)

    Because :math:`A - A^T` is skew-symmetric, :math:`R` is orthogonal.

    Arguments:
        A: A square matrix :math:`A`, with shape :math:`(*, D, D)`.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, A: Tensor, **kwargs):
        super().__init__(**kwargs)

        self.R = torch.linalg.matrix_exp(A - A.mT)

    def _call(self, x: Tensor) -> Tensor:
        return torch.einsum("...ij,...j->...i", self.R, x)

    def _inverse(self, y: Tensor) -> Tensor:
        return torch.einsum("...ij,...i->...j", self.R, y)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.zeros_like(x[..., 0])


class LULinearTransform(Transform):
    r"""Creates a linear transformation :math:`f(x) = L U x`.

    Arguments:
        LU: A matrix whose lower and upper triangular parts are the non-zero elements
            of :math:`L` and :math:`U`, with shape :math:`(*, D, D)`.
    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, LU: Tensor, **kwargs):
        super().__init__(**kwargs)

        I = torch.eye(LU.shape[-1], dtype=LU.dtype, device=LU.device)

        self.L = torch.tril(LU)
        self.U = torch.triu(LU, diagonal=1) + I

    def _call(self, x: Tensor) -> Tensor:
        return torch.einsum("...ij,...j->...i", self.L @ self.U, x)

    def _inverse(self, y: Tensor) -> Tensor:
        return torch.linalg.solve_triangular(
            self.U,
            torch.linalg.solve_triangular(
                self.L,
                y.unsqueeze(-1),
                upper=False,
                unitriangular=False,
            ),
            upper=True,
            unitriangular=True,
        ).squeeze(-1)

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        diag = torch.diagonal(self.L, dim1=-1, dim2=-2)
        ladj = diag.abs().log().sum(dim=-1)

        return ladj.expand_as(x[..., 0])



r"""Bayesian models."""

from __future__ import annotations

__all__ = [
    "BayesianModel",
]

import copy
import math
import re
import torch
import torch.nn as nn

from contextlib import contextmanager
from functools import partial
from torch import Tensor
from typing import Dict, Generator, Sequence

from .nn import Linear, MaskedLinear, linear


def _compile(prefix: str) -> re.Pattern:
    assert re.fullmatch(r"[\w\.\*]*", prefix) is not None, f"Invalid prefix {prefix}."

    # fmt: off
    pattern = (
        prefix
        .replace(".", r"\.")
        .replace("**", r"[\w\.]+")
        .replace("*", r"\w+")
        + r".*"
    )
    # fmt: on

    return re.compile(pattern)


def _match(pattern: re.Pattern, string: str) -> bool:
    return re.fullmatch(pattern, string) is not None


def _softclip(x: Tensor, bound: float) -> Tensor:
    return x * (1 + (x / bound).square()).rsqrt()


class BayesianModel(nn.Module):
    r"""Creates a Bayesian wrapper around a base model.

    Arguments:
        base: A base model.
        include_params: A list of parameter name prefixes to include. In a prefix, a
            single star `*` matches alphanumeric strings (`[a-zA-Z0-9_]+`) and a double
            star `**` matches dot-separated alphanumeric strings (`[a-zA-Z0-9_\.]+`). By
            default, all parameters are included.
        exclude_params: A list of parameter name prefixes to exclude.
    """

    def __init__(
        self,
        base: nn.Module,
        init_logvar: float = -9.0,
        include_params: Sequence[str] = ("",),
        exclude_params: Sequence[str] = (),
    ):
        super().__init__()

        self.base = base

        self.means = nn.ParameterDict()
        self.logvars = nn.ParameterDict()

        include_patterns = [_compile(prefix) for prefix in include_params]
        exclude_patterns = [_compile(prefix) for prefix in exclude_params]

        for name, param in self.base.named_parameters():
            if not any(_match(pattern, name) for pattern in include_patterns):
                continue
            elif any(_match(pattern, name) for pattern in exclude_patterns):
                continue

            key = name.replace(".", "-")

            self.means[key] = nn.Parameter(torch.empty_like(param))
            self.logvars[key] = nn.Parameter(torch.empty_like(param))

        self._reset_bayesian_parameters(logvar_mean=init_logvar)

    def _reset_bayesian_parameters(self, logvar_mean: float = -9.0, logvar_std: float = 1e-3):
        r"""Initializes the posterior means and log-variances."""

        for name, param in self.base.named_parameters():
            key = name.replace(".", "-")

            if key in self.means:
                self.means[key].data.copy_(param.data)
                self.logvars[key].data.normal_(logvar_mean, logvar_std)

    def forward(self, *args, **kwargs):
        raise RuntimeError(
            "BayesianModel should not be called directly. Use `sample_model` or `reparameterize` instead."
        )

    def sample_params(self) -> Dict[str, Tensor]:
        r"""Returns model parameters sampled from the posterior."""

        params = {}

        for key in self.means.keys():
            name = key.replace("-", ".")

            mean = self.means[key]
            std = _softclip(self.logvars[key], 18.0).div(2.0).exp()

            params[name] = mean + std * torch.randn_like(mean)

        return params

    def sample_model(self) -> nn.Module:
        r"""Returns a model sampled from the posterior.

        The returned model is a standalone copy of the base model where parameters have
        been replaced. It can be used exactly in the same ways as the base model.

        Warning:
            This method should not be used for training as it cannot propagate gradients
            to the Bayesian model's internal parameters.
        """

        with torch.no_grad():
            params = self.sample_params()

        new_base = copy.deepcopy(self.base)
        new_base.load_state_dict(params, strict=False)

        return new_base

    @contextmanager
    def reparameterize(self, local_trick: bool = False) -> Generator[nn.Module]:
        r"""Reparameterizes the base model from the posterior.

        Within this context, the base model is temporarily reparametrized and behaves
        deterministically, meaning that two calls to a method with the same inputs,
        leads to the same outputs.

        Arguments:
            local_trick: Whether to use the local reparameterization trick for linear
                layers.

        Yields:
            The reparametrized base model.

        References:
            | Variational Dropout and the Local Reparameterization Trick (Kingma et al., 2015)
            | https://arxiv.org/abs/1506.02557
        """

        params = self.sample_params()

        with torch.nn.utils.stateless._reparametrize_module(self.base, params):
            if local_trick:
                with self._reparametrize_trick():
                    yield self.base
            else:
                yield self.base

    @contextmanager
    def _reparametrize_trick(self):
        original_forwards = {}
        randn_cache = {}

        for name, module in self.base.named_modules():
            key = name.replace(".", "-")

            if type(module) not in (nn.Linear, Linear, MaskedLinear):
                continue
            elif key + "-weight" not in self.means:
                continue

            original_forwards[name] = module.forward
            module.forward = partial(
                self._local_reparameterization_trick,
                name=name,
                module=module,
                randn_cache=randn_cache,
            )

        try:
            yield
        finally:
            for name, module in self.base.named_modules():
                if name in original_forwards:
                    module.forward = original_forwards[name]

    def _local_reparameterization_trick(
        self,
        x: Tensor,
        name: str,
        module: nn.Module,
        randn_cache: Dict[(str, torch.Size), Tensor],
    ) -> Tensor:
        """Applies the local reparameterization trick to a linear layer."""

        key = name.replace(".", "-")

        w_mean = self.means[key + "-weight"]
        w_var = _softclip(self.logvars[key + "-weight"], 18.0).exp()

        if isinstance(module, MaskedLinear):
            w_mean = module.mask * w_mean
            w_var = module.mask * w_var

        if key + "-bias" in self.means:
            b_mean = self.means[key + "-bias"]
            b_var = _softclip(self.logvars[key + "-bias"], 18.0).exp()
        else:
            b_mean = module.bias
            b_var = None

        y_mean = linear(x, w_mean, b_mean)
        y_var = linear(x.square(), w_var, b_var)

        address = (name, y_mean.shape)

        if address in randn_cache:
            eta = randn_cache[address]
        else:
            eta = randn_cache[address] = torch.randn(
                y_mean.shape,
                dtype=y_mean.dtype,
                device=y_mean.device,
            )

        y = y_mean + y_var.sqrt() * eta

        return y

    def kl_divergence(self, prior_var: float = 1.0) -> Tensor:
        r"""Returns the KL divergence between the posterior and the prior."""

        kl = 0.0

        for key in self.means.keys():
            mean = self.means[key]
            log_var = _softclip(self.logvars[key], 18.0)
            var = log_var.exp()

            # fmt: off
            kl = kl + 0.5 * (
                var / prior_var
                + mean.square() / prior_var
                + math.log(prior_var) - log_var
                - 1.0
            ).sum()
            # fmt: on

        return kl



r"""Tests for the zuko.transforms module."""

import pytest
import torch

from torch import randn
from torch.distributions import *

from zuko.transforms import *


@pytest.mark.parametrize("batched", [False, True])
def test_univariate_transforms(batched: bool):
    batch = (256,) if batched else ()

    ts = [
        IdentityTransform(),
        CosTransform(),
        SinTransform(),
        SoftclipTransform(),
        CircularShiftTransform(),
        SignedPowerTransform(randn(batch)),
        AdditiveTransform(randn(batch)),
        MonotonicAffineTransform(randn(batch), randn(batch)),
        MonotonicRQSTransform(randn(*batch, 8), randn(*batch, 8), randn(*batch, 7)),
        MonotonicTransform(lambda x: x**3),
        BernsteinTransform(randn(*batch, 16)),
        BoundedBernsteinTransform(randn(*batch, 16)),
        GaussianizationTransform(randn(*batch, 8), randn(*batch, 8)),
        UnconstrainedMonotonicTransform(lambda x: torch.exp(-(x**2)) + 1e-2),
        SOSPolynomialTransform(randn(*batch, 3, 5)),
    ]

    for t in ts:
        # Call
        if hasattr(t.domain, "lower_bound"):
            x = torch.linspace(t.domain.lower_bound + 1e-2, t.domain.upper_bound - 1e-2, 256)
        else:
            x = torch.linspace(-5.0, 5.0, 256)

        y = t(x)

        assert x.shape == y.shape, t

        # Inverse
        z = t.inv(y)

        assert torch.allclose(x, z, atol=1e-4), t

        # Jacobian
        J = torch.autograd.functional.jacobian(t, x)

        assert (torch.triu(J, diagonal=+1) == 0).all(), t
        assert (torch.tril(J, diagonal=-1) == 0).all(), t

        ladj = torch.diag(J).abs().log()

        assert torch.allclose(t.log_abs_det_jacobian(x, y), ladj, atol=1e-4), t

        # Compound call
        y_comp, ladj_comp = t.call_and_ladj(x)

        assert torch.allclose(y_comp, y, atol=1e-4), t
        assert torch.allclose(ladj_comp, ladj, atol=1e-4), t

        # Inverse Jacobian
        J = torch.autograd.functional.jacobian(t.inv, y)

        assert (torch.triu(J, diagonal=+1) == 0).all(), t
        assert (torch.tril(J, diagonal=-1) == 0).all(), t

        ladj = torch.diag(J).abs().log()

        assert torch.allclose(t.inv.log_abs_det_jacobian(y, z), ladj, atol=1e-4), t


def test_multivariate_transforms():
    A, B = torch.randn(5, 16), torch.randn(16, 5)
    f = lambda t, x: torch.sigmoid(x @ A) @ B

    ts = [
        FreeFormJacobianTransform(f, 0.0, 1.0, atol=1e-7, rtol=1e-6),
        PermutationTransform(torch.randperm(5)),
        RotationTransform(randn(5, 5)),
        LULinearTransform(randn(5, 5)),
    ]

    for t in ts:
        # Shapes
        x = randn(256, 5)
        y = t(x)

        assert x.shape == y.shape, t

        # Inverse
        z = t.inv(y)

        assert torch.allclose(x, z, atol=1e-4), t

        # Jacobian
        x = randn(5)
        y = t(x)

        J = torch.autograd.functional.jacobian(t, x)
        ladj = torch.linalg.slogdet(J).logabsdet

        assert torch.allclose(t.log_abs_det_jacobian(x, y), ladj, atol=1e-4), t

        # Compound call
        y_comp, ladj_comp = t.call_and_ladj(x)

        assert torch.allclose(y_comp, y, atol=1e-4), t
        assert torch.allclose(ladj_comp, ladj, atol=1e-4), t

        # Inverse Jacobian
        z = t.inv(y)

        J = torch.autograd.functional.jacobian(t.inv, y)
        ladj = torch.linalg.slogdet(J).logabsdet

        assert torch.allclose(t.inv.log_abs_det_jacobian(y, z), ladj, atol=1e-4), t



r"""Tests for the zuko.mixtures module."""

import pytest
import torch

from pathlib import Path
from torch import randn
from typing import Sequence

from zuko.mixtures import *


@pytest.mark.parametrize("M", [GMM])
def test_mixtures(tmp_path: Path, M: type):
    mixture = M(3, 5)

    # Evaluation of log_prob
    x, c = randn(256, 3), randn(5)
    log_p = mixture(c).log_prob(x)

    assert log_p.shape == (256,)
    assert log_p.requires_grad

    mixture.zero_grad(set_to_none=True)
    loss = -log_p.mean()
    loss.backward()

    for p in mixture.parameters():
        assert p.grad is not None

    # Sampling
    x = mixture(c).sample((32,))

    assert x.shape == (32, 3)

    # Reparameterization trick
    if mixture(c).has_rsample:
        x = mixture(c).rsample()

        mixture.zero_grad(set_to_none=True)
        loss = x.square().sum().sqrt()
        loss.backward()

        for p in mixture.parameters():
            assert p.grad is not None

    # Saving
    torch.save(mixture, tmp_path / "mixture.pth")

    # Loading
    mixture_bis = torch.load(tmp_path / "mixture.pth", weights_only=False)

    x, c = randn(3), randn(5)

    with torch.random.fork_rng():
        log_p = mixture(c).log_prob(x)

    with torch.random.fork_rng():
        log_p_bis = mixture_bis(c).log_prob(x)

    assert torch.allclose(log_p, log_p_bis)

    # Printing
    assert repr(mixture)


@pytest.mark.parametrize("batch", [(), (4,)])
@pytest.mark.parametrize("features", [3])
@pytest.mark.parametrize("context", [0, 5])
@pytest.mark.parametrize("components", [2])
@pytest.mark.parametrize("covariance_type", ["full", "diagonal", "spherical"])
@pytest.mark.parametrize("tied", [False, True])
@pytest.mark.parametrize("strategy", [None, "random", "kmeans", "kmeans++"])
def test_gmm_shapes(
    batch: Sequence[int],
    features: int,
    context: int,
    components: int,
    covariance_type: str,
    tied: bool,
    strategy: str,
):
    if context > 0:
        c = torch.randn(*batch, context)
    else:
        c, batch = None, ()

    gmm = GMM(
        features=features,
        context=context,
        components=components,
        covariance_type=covariance_type,
        tied=tied,
    )

    # Initialization
    if strategy is not None:
        gmm.initialize(torch.randn(1024, features), strategy=strategy)

    # Forward
    d = gmm(c)

    # Shapes
    assert d.batch_shape == (*batch,)
    assert d.event_shape == (features,)

    assert d.logits.shape == (*batch, components)

    if covariance_type == "full":
        assert d.base.loc.shape == (*batch, components, features)
        assert d.base.covariance_matrix.shape == (*batch, components, features, features)
    else:
        assert d.base.base_dist.loc.shape == (*batch, components, features)
        assert d.base.base_dist.scale.shape == (*batch, components, features)


@pytest.mark.parametrize("covariance_type", ["full", "diagonal", "spherical"])
def test_gmm_tied_covariance(covariance_type: str):
    gmm = GMM(features=3, components=2, covariance_type=covariance_type, tied=True)
    d = gmm()

    if covariance_type == "full":
        covs = d.base.covariance_matrix
    else:
        covs = d.base.base_dist.scale

    assert torch.allclose(covs[0], covs[1])


def test_gmm_insufficient_samples():
    gmm = GMM(features=3, components=7, covariance_type="full")

    with pytest.raises(AssertionError, match="The number of samples"):
        gmm.initialize(torch.randn(6, 3), strategy="random")
